import re

import sastadev.CHAT_Annotation as sastachat
#from CHAT_Annotation import CHAT_patterns, interpunction, wordpat
from sastadev.sastatoken import stringlist2tokenlist


def alts(pats, grouping=False):
    result = r''
    for pat in pats:
        if result == r'':
            result = pat
        else:
            result += '|' + pat
    if grouping:
        result = '(' + result + ')'
    else:
        result = '(?:' + result + ')'
    return result


#interpunction = r'[!\?\.,;]'  # add colon separated by spaces
#word = r'[^!\?;\.\[\]<>\s]+'
word = sastachat.wordpat
scope = r'<.+?>'
scopeorword = alts([scope, word])
myrepetition = r'\[x\s*[0-9]+\s*\]'
replacement = scopeorword + r'\s*\[:.+?\]'
realwordreplacement = scopeorword + r'\s*\[::.+?\]'
alternativetranscription = r'\[=\?.+?\]'
# dependenttier # p. 71
commentonmainline = r'\[%.+?\]'  # p. 71
#bestguess = scopeorword+r'\s*\[\?\]' #p. 70-71
bestguess = r'\[\?\]'
#overlap follows p. 71
#overlap precedes p. 71
repetition = scopeorword + r'\s*\[/\]\s*' + word  # p73 should actually cover the number of words inside the scope
retracing = scopeorword + r'\s*\[//\]\s*' + word  # p73
whitespace = r'\s+'


#sastaspecials = [r'\[::', r'\[=', r'\[:', r'\[=\?', r'\[x', r'\<', r'\>', r'\[\?\]', r'\[/\]', r'\[//\]', r'\[///\]', r'\[%', r'\]']
sastaspecials = list(sastachat.CHAT_patterns)
sastapatterns = sorted(sastaspecials, key=lambda x: len(x), reverse=True) + [word, sastachat.interpunction]
fullsastapatterns = alts(sastapatterns)
fullsastare = re.compile(fullsastapatterns)

allpatterns = [realwordreplacement, replacement, myrepetition, alternativetranscription, commentonmainline, bestguess, retracing]
sortedallpatterns = sorted(allpatterns, key=lambda x: len(x), reverse=True) + [word, sastachat.interpunction]
fullpattern = alts(sortedallpatterns)
#print(fullpattern)
fullre = re.compile(fullpattern)


def tokenize(instring):
    tokenstring = fullre.findall(instring)
    result = stringlist2tokenlist(tokenstring)
    return result


def sasta_tokenize(instring):
    if instring is None:
        return []
    tokenstring = fullsastare.findall(instring)
    result = stringlist2tokenlist(tokenstring, start=10, inc=10)
    return result



from typing import List, Set

from sastadev.sastatypes import SynTree
from sastadev.Sziplus import getnodecount
from sastadev.treebankfunctions import getattval

noimplemmas = {'hoeven', 'moeten', 'mogen', 'kunnen', 'hebben', 'willen', 'hebben', 'zitten'}
noimpwords = {'ben', 'bent', 'is', 'zijn'}
impmodlemmas = {'eens', 'maar'}


impquery = '''
.//node[@cat="sv1" and
    not(node[@rel="su"]) and
    node[@rel="hd" and @pt="ww" and @pvtijd="tgw" and @pvagr="ev"
         and @wvorm="pv" and
         not(contains(@frame,"modal")) and
         (not(@lemma="zijn") or @word="wees" or @word="weest") and
         (not(contains(@lemma,"moeten") or contains(@lemma,"hoeven") or
              contains(@lemma,"zullen") or contains(@lemma,"kunnen") or
              contains(@lemma,"mogen") or @lemma="hebben" or contains(@lemma, "_hebben")
             )
         )
        ]
     ]
'''

#//node[@cat="sv1" and
#    count(node)<=2 and
#    (not(node[@rel="su"]) or (node[@rel="su" and @lemma="jij"])) and
#    node[@rel="hd" and @pt="ww" and @pvtijd="tgw" and @pvagr="ev"
#         and @wvorm="pv" and
#         not(contains(@frame,"modal")) and
#         @lemma!="hoeven" and @lemma!="moeten" and @lemma!="kunnen" and @lemma !="mogen" and @lemma!="hebben" and @lemma!="zitten" and @word!="ben" and
#         (@stype="imparative" or ../node[@rel="mod" and (@lemma="eens" or @lemma="maar")])  ] ]


ynquery = '''
.//node[@cat="sv1" and
    node[@rel="hd" and @pt="ww" and @pvtijd !="conj" and @stype="ynquestion" ] and
    node[@rel="su" ]
 ]
'''


'''
//node[@cat="sv1" and
    node[@rel="hd" and @pt="ww" and @pvtijd !="conj" and @stype="ynquestion" and number(@begin) < ../node[@rel="su" ]/number(@begin)] and
    node[@rel="su" and @lemma!="ik"]  and
    not(node[@rel="vc"]) and
    not(node[@rel="mod" and (@lemma="maar" or @lemma="eens")]) and
    count(node) <=3]
'''


def ynwi(syntree, ncf):
    results = []
    cands = syntree.xpath(ynquery)
    #print('#cands', len(cands))
    ok = True
    for cand in cands:
        hds = cand.xpath('node[@rel="hd"]')
        sus = cand.xpath('node[@rel="su"]')
        modlemmas = set(cand.xpath('node[@rel="mod"]/@lemma'))

        for hd in hds:
            for su in sus:
                suend = getattval(su, 'end')
                hdend = getattval(hd, 'end')
                hdpvtijd = getattval(hd, 'pvtijd')
                hdpvagr = getattval(hd, 'pvagr')
                hdlemma = getattval(hd, 'lemma')
                hdword = getattval(hd, 'words')
                hdpotimp = (hdpvtijd == 'tgw' and hdpvagr == 'ev' and (hdlemma != 'zijn' or hdword == 'wees'))
                ok = ok and (hdend < suend)
                #print('ok=', ok)
                #noimpmodlemmas = modlemmas.intersection(impmodlemmas)
                #print('noimpmodlemmas', noimpmodlemmas)
                #ok = ok and ((not hdpotimp) or noimpmodlemmas == {})
                #print ('hdpotimp',hdpotimp)
                #print('modlemmas', modlemmas)
                #print('ok=', ok)
                thenodecount = getnodecount(cand)
                #print('thenodecount', thenodecount)
                ok = ok and ncf(thenodecount)
                #print('ok=', ok)
                if ok:
                    results.append(cand)
    return results


def wondx(syntree):
    results = ynwi(syntree, lambda x: 2 <= x <= 3)
    return results


def wond4(syntree):
    results = ynwi(syntree, lambda x: x == 4)
    return results


def wond5plus(syntree):
    results = ynwi(syntree, lambda x: x >= 5)
    return results


def impwi(syntree: SynTree, nodecounts: Set[int]) -> List[SynTree]:
    '''
    The function impwi finds nodes for imperative clauses in *syntree* with the number of nodes equal to one of the  integers in the *nodecounts* set
    '''
    results = []
    cands = syntree.xpath(impquery)
    ok = True
    for cand in cands:
        hds = cand.xpath('node[@rel="hd"]')
        modlemmas = set(cand.xpath('node[@rel="mod"]/@lemma'))
        #print('modlemmas', modlemmas)

        for hd in hds:
            lemma = getattval(hd, 'lemma')
            ok = ok and (lemma not in noimplemmas)
            #print('ok=', ok)
            word = getattval(hd, 'word')
            ok = ok and (word not in noimpwords)
            #print('ok=', ok)
            stype = getattval(hd, 'stype')
            ok = ok and (stype == 'imparative' or (modlemmas.intersection(impmodlemmas) != {}))
            #print('ok=', ok)

        thenodecount = getnodecount(cand)
        ok = ok and thenodecount in nodecounts
        #print('ok=', ok)
        if ok:
            results.append(cand)
    return results


def wx(syntree: SynTree) -> List[SynTree]:
    '''
    The function *wx* finds nodes for imperative clauses with 1 or 2 nodes.
    it uses the function *impwi* to achieve that.
    '''
    results = impwi(syntree, {1, 2})
    return results


def wxy(syntree):
    results = impwi(syntree, {3})
    return results


def wxyz(syntree):
    results = impwi(syntree, {4})
    return results


def wxyz5(syntree):
    results = impwi(syntree, {5})
    return results



from collections import defaultdict
from typing import Dict, List, Tuple

from sastadev.deregularise import correctinflection
from sastadev.metadata import bpl_word, defaultpenalty
from sastadev.sastatoken import Token
from sastadev.sastatypes import ReplacementMode, SynTree, TokenTreePredicate
from sastadev.treebankfunctions import find1

BasicExpansion = Tuple[str, List[str], str, str, str, int]
BasicReplacement = Tuple[str, str, str, str, str, int]
KnownReplacement = Tuple[str, str, str, str, str, ReplacementMode]

dp = defaultpenalty
dp6 = dp + 6
dp3 = dp + 3
dp2 = dp + 2
pron = 'Pronunciation'
orth = 'Orthography'
infpron = 'Informal Pronunciation'
initdev = 'Initial Devoicing'
codared = 'Coda Reduction'
apomiss = 'Missing Apostrophe'
addschwa = 'Schwa addition'
onsetred = 'Onset Reduction'
d_er = 'd-onset on er'
spellerr = 'Spelling Error'
varpron = 'Alternative Pronunciation'
pronerr = 'Wrong Pronunciation'
contract = 'Contraction'
t_ie = 't-Insertion with ie'
fndrop = 'Final n drop'
zdev = 'Devoicing of /z/'
wrongpron = 'Wrong Prunciation'
phonrepl = '/{wrong}/ instead of /{correct}/'
wronginfl = 'Incorrect inflection'
morph = 'Morphology'
overgen = 'Overgeneralisation'
typo = 'Typo'
typorepl = '{wrong} instead of {correct}'
redpron = 'reduced pronunciation'
emphasis = 'emphasis'
prefixdrop = 'prefix drop'
voweldup = 'vowel duplication'
erdrop = 'er drop'
closedvow = 'closed vowel'
vzdevoicing = 'v/z devoicing'

#: The constant *Rvzlist* contains a list of adpositions that can combine with *er* into
#: a so-called pronominal adverb. Example: *van*, since *er* + * van* = *ervan*.
#:
Rvzlist = ['aan', 'achter', 'achteraan', 'achterin', 'achterop', 'af', 'beneden', 'benevens', 'bij', 'binnen',
           'binnenuit', 'boven', 'bovenaan', 'bovenin', 'bovenop', 'buiten', 'dichtbij', 'door', 'doorheen', 'heen',
           'in', 'langs', 'langsheen', 'langszij', 'mede', 'mee', 'middendoor', 'na', 'naar', 'naartoe', 'naast',
           'nabij', 'neer', 'om', 'omheen', 'onder', 'onderaan', 'op', 'over', 'overheen', 'rond', 'rondom', 'tegen',
           'tegenaan', 'tegenin', 'tegenover', 'toe', 'tussen', 'tussenin', 'tusssen', 'uit', 'van', 'vanaf',
           'vandaan', 'vandoor', 'vanop', 'vanuit', 'vanwege', 'vlakbij', 'voor', 'vooraan', 'voorbij', 'zonder']

#: The constant *ervzvariants* contains a list of 6-tuples (type *BasicReplacement*)
#: for variants of *er* (*der*, *d'r*) plus an adposition. The list is generated by using
#: the constant *Rvzlist*
#:
#: .. autodata:: basicreplacements::Rvzlist
#:        :no-value:
#:
ervzvariants: List[BasicReplacement] = \
    [('der' + vz, 'er' + vz, pron, varpron, d_er, dp) for vz in Rvzlist] + \
    [("d'r" + vz, 'er' + vz, pron, varpron, d_er, dp) for vz in Rvzlist]

#: The constant *innereplacements* contains replacements for short closed syllable prepositions followed by *e*
#: (and with final consonant doubling), e.g. *inne* for the preposition *in*. They can be replaced by *in* or by *erin*.
#: The preferred replacement for *inne* and *innu* is by an expansion into preposition + definite article, so
#: replacement by *in* has a higher penalty than such an expansion, and replacement by *erin* has an even higher penalty.
#:
innereplacements = [('inne', 'erin', pron, wrongpron, erdrop, dp6),
                    ('inne', 'in', pron, wrongpron, emphasis, dp3),
                    ('omme', 'erom', pron, wrongpron, erdrop, dp6),
                    ('omme', 'om', pron, wrongpron, emphasis, dp3),
                    ('oppe', 'erop', pron, wrongpron, erdrop, dp6),
                    ('oppe', 'op', pron, wrongpron, emphasis, dp3),
                    ('vanne', 'ervan', pron, wrongpron, erdrop, dp6),
                    ('vanne', 'van', pron, wrongpron, emphasis, dp3),
                    ]
#: The constant *innureplacements* is derived from the constant *innereplacements* by replacing the final *e* of the
#: first word in each tuple by an *u*.
#:
innureplacements = [(w1[:-1] + 'u', w2, c, n, v, p) for (w1, w2, c, n, v, p) in innereplacements]

#: The constant *basicreplacementlist* contains 6-tuples of strings with:
#:
#: * frequent deviant spelling of a word, usually to indicate how the word was (incorrectly) pronounced. together with (e.g. "as", "effe")
#: * the correct spelling of the word ("als", "even")
#: * linguistic level (e.g Pronunciation, Orthography)
#: * characterisation of the deviancy, e.g. "Informal Pronunciation", "Wrong Pronunciation"
#: * details about the deviation, e.g. "Coda Reduction", "Initial Devoicing"
#: * penalty for the replacement
#:
#: It also includes variants of R-pronouns + adposition, e.g. *dervan* or *d'rvan*
#: instead of *ervan* for all adpositions that allow an R-pronoun. It derives these by
#: using the constant *ervzvariants*
#:
#: .. autodata:: basicreplacements::ervzvariants
#:       :no-value:
#:
#: It obtains replacements for short prepositions with 'e' or 'u' attached (e.g. *inne*, *innu* for *in*) from the constants
#: *innereplacements* and *innureplacements*.
#:
#: .. autodata:: basicreplacements::innereplacements
#:      :no-value:
#:
#: .. autodata:: basicreplacements::innureplacements
#:      :no-value:
#:
basicreplacementlist: List[BasicReplacement] = [('as', 'als', pron, infpron, codared, dp),
                                                ('isse', 'is', pron, infpron, addschwa, dp),
                                                ('ooke', 'ook', pron, infpron, addschwa, dp),
                                                ('t', "'t", orth, spellerr, apomiss, dp),
                                                ('effjes', 'eventjes', pron, infpron, varpron, dp),
                                                ('effetjes', 'eventjes', pron, infpron, varpron, dp),
                                                ('effe', 'even', pron, infpron, varpron, dp),
                                                ('set', 'zet', pron, infpron, initdev, dp),
                                                ('fan', 'van', pron, infpron, initdev, dp),
                                                ('hie', 'hier', pron, pronerr, codared, dp),
                                                ('eers', 'eerst', pron, pronerr, codared, dp),
                                                ('era', 'eraf', pron, pronerr, codared, dp),
                                                ('il', 'wil', pron, pronerr, onsetred, dp),
                                                ('tee', 'twee', pron, pronerr, onsetred, dp),
                                                ('nie', 'niet', pron, infpron, codared, dp),
                                                ('s', 'is', orth, spellerr, apomiss, dp),
                                                ('ooke', 'ook', pron, infpron, addschwa, dp),
                                                ('it', 'dit', pron, pronerr, onsetred, dp),
                                                ('da', 'dat', pron, infpron, codared, dp),
                                                ('si', 'zit', pron, infpron, codared, dp),  # and zdev
                                                ('ieduleen', 'iedereen', pron, wrongpron, phonrepl.format(wrong='l', correct='r'), dp),
                                                ('allemaaw', 'allemaal', pron, wrongpron, phonrepl.format(wrong='w', correct='l'), dp),
                                                ('amaal', 'allemaal', pron, infpron, varpron, dp),
                                                ('wiw', 'wil', pron, wrongpron, phonrepl.format(wrong='w', correct='l'), dp),
                                                ('annug', 'ander', pron, wrongpron, phonrepl.format(wrong='nug', correct='der'), dp),
                                                ('nohug', 'nodig', pron, wrongpron, phonrepl.format(wrong='hu', correct='di'), dp),
                                                ('magge', 'mogen', morph, wronginfl, '{} & {}'.format(overgen, infpron), dp),
                                                ('maggen', 'mogen', morph, wronginfl, overgen, dp),
                                                ('aleen', 'alleen', orth, typo, typorepl.format(wrong='alleen', correct='alleen'), dp),
                                                ('heef', 'heeft', pron, infpron, codared, dp),
                                                ('saan', 'staan', pron, wrongpron, onsetred, dp),
                                                ('saan', 'gaan', pron, wrongpron, wrongpron, dp + 2),
                                                ('jerke', 'werken', pron, wrongpron, wrongpron, dp),
                                                ('taan', 'staan', pron, wrongpron, onsetred, dp),
                                                ("a'maal", 'allemaal', pron, infpron, redpron, dp),
                                                ('taan', 'staan', pron, wrongpron, onsetred, dp),
                                                ('beurt', 'gebeurt', pron, wrongpron, prefixdrop, dp),
                                                ('dahaar', 'daar', pron, emphasis, voweldup, dp),
                                                ('desu', 'deze', pron, infpron, vzdevoicing, dp),
                                                ('tan', 'dan', pron, infpron, initdev, dp),
                                                ('tat', 'dat', pron, infpron, initdev, dp),
                                                ('tit', 'dit', pron, infpron, initdev, dp),
                                                ('lape', 'slapen', pron, infpron, f'{onsetred}+{fndrop}', dp),
                                                ('vas', 'vast', pron, infpron, codared, dp),
                                                ('datte', 'dat', pron, infpron, emphasis, dp),
                                                ('omdatte', 'dat', pron, infpron, emphasis, dp),
                                                ('cirtus', 'circus', pron, wrongpron, typorepl.format(wrong='t', correct='c'), dp)

                                                ] + \
    ervzvariants + \
    innereplacements + \
    innureplacements


#: The dictionary *basicreplacements* maps a word with deviant orthography to a list of
#: 5-tuples containing the correct orthography, 3 strings to encode metadata, and an integer for the associated penalty. The
#: dictionary is derived from the constant *basicreplacementlist*
#:
#: .. autodata:: basicreplacements::basicreplacementlist
#:       :no-value:
#:
basicreplacements: Dict[str, List[Tuple[List[str], str, str, str, int]]] = defaultdict(list)
for w1, w2, c, n, v, p in basicreplacementlist:
    basicreplacements[w1].append((w2, c, n, v, p))

#closedsyllshortpreps = ['in', 'met',  'om',  'op',  'van' ]  # langs removed
#: The constant *closedsyllshortexpansion* contains expansions for short closed syllable prepositions followed by *e*
#: (and with final consonant doubling), e.g. *inne* for the preposition *in*. They can be expanded to preposition + *de*
#: (e.g. *in de*) or preposition + *het* (e.g. *in het*), with a preference for the former, indicated by a lower
#:penalty than the second alternative expansion.
#:
closesyllshortprepexpansions = [('inne', ['in', 'de'], pron, infpron, contract, dp),
                                ('inne', ['in', 'het'], pron, infpron, contract, dp2),  # het-cases temporarily put off, getexpansions contains a bugf
                                ('mette', ['met', 'de'], pron, infpron, contract, dp),
                                ('mette', ['met', 'het'], pron, infpron, contract, dp2),
                                ('omme', ['om', 'de'], pron, infpron, contract, dp),
                                ('omme', ['om', 'het'], pron, infpron, contract, dp2),
                                ('oppe', ['op', 'de'], pron, infpron, contract, dp),
                                ('oppe', ['op', 'het'], pron, infpron, contract, dp2),
                                ('vanne', ['van', 'de'], pron, infpron, contract, dp),
                                ('vanne', ['van', 'het'], pron, infpron, contract, dp2),
                                ]

#: The constant *innuclosedsyllshortexpansion* contains expansions for short closed syllable prepositions followed by *u*
#: (and with final consonant doubling), e.g. *innu* for the preposition *in*. It is derived from the constant
#: *closedsyllshortprepexpansions* by replacing the final *e* of the first word of the tuple by *u*.
#:
innuclosedsyllshortprepexpansions = [(w1[:-1] + 'u', w2, c, n, v, p)
                                     for (w1, w2, c, n, v, p) in closesyllshortprepexpansions]

#opensyllshortpreps = ['aan',  'door',  'uit',   'voor', ]

#: The constant *basicexpansionlist* is the concatenation of a number of lists of type *BasicExpansion*.
#:
#: * A list of  6-tuples  with:
#:
#:   * frequent contracted spelling of a word sequence, usually to indicate how the word was actually (possibly incorrectly) pronounced, (e.g. "innu", "das")
#:   * the correct expansion of the word as a list of strings (["in", "de"], ["dat", "is"])
#:   * linguistic level (e.g "Pronunciation", "Orthography")
#:   * characterisation of the deviancy, e.g. "Informal Pronunciation"
#:   * details about the deviancy, e.g. "Contraction"
#:   * penalty for this modification
#:
#: * the constant *closesyllshortprepexpansions*:
#:   * .. autodata:: basicreplacements::closesyllshortprepexpansions
#:
#: * the constant *innuclosesyllshortprepexpansions*:
#:   * .. autodata:: basicreplacements::innuclosesyllshortprepexpansions
#:
basicexpansionlist: List[BasicExpansion] = \
    [('dis', ['dit', 'is'], pron, infpron, contract, dp),
     ('das', ['dat', 'is'], pron, infpron, contract, dp),
     ('tis', ['dit', 'is'], pron, infpron, contract, dp),
     ('waas', ['waar', 'is'], pron, infpron, contract, dp),
     ('is-t-ie', ['is', 'ie'], pron, infpron, t_ie, dp),
     ('als-t-ie', ['als', 'ie'], pron, infpron, t_ie, dp),
     ('of-t-ie', ['of', 'ie'], pron, infpron, t_ie, dp),
     ('as-t-ie', ['als', 'ie'], pron, infpron, t_ie, dp),
     ("dit's", ["dit", "is"], pron, infpron, contract, dp),
     ("dat's", ["dat", "is"], pron, infpron, contract, dp),
     ("datte", ['dat', 'ie'], pron, infpron, contract, dp + 2),
     ("omdatte", ['omdat', 'ie'], pron, infpron, contract, dp + 2)
     ]
# + closesyllshortprepexpansions # put off does not lead to improvement
# + innuclosedsyllshortprepexpansions # put off does not lead to improvement


#: The dictionary *basicexpansions* maps a contracted word form to a list of 4-tuples
#: containing its expansion as a list of words, plus 3 strings to encode metadata. The
#: dictionary is derived from the constant *basicexpansionlist*:
#:
#: .. autodata:: basicreplacements::basicexpansionlist
#:       :no-value:
#:
basicexpansions: Dict[str, List[Tuple[List[str], str, str, str, int]]] = defaultdict(list)
for (w1, w2, c, n, v, p) in basicexpansionlist:
    basicexpansions[w1].append((w2, c, n, v, p))

knownreplacements: List[KnownReplacement] = [
    ('ze', "z'n", pron, infpron, fndrop, bpl_word),
    ('desu', 'deze', pron, infpron, zdev, bpl_word),
    ('mij', 'mijn', pron, infpron, fndrop, bpl_word),

]

knownreplacementsdict: Dict[Tuple[str, str], KnownReplacement] = {(repl[0], repl[1]): repl for repl in knownreplacements}


def getmeta4CHATreplacements(wrongword: str, correctword: str) -> KnownReplacement:
    if (wrongword, correctword) in knownreplacementsdict:
        result = knownreplacementsdict[(wrongword, correctword)]
    else:
        nwms = correctinflection(wrongword)
        if nwms != []:
            for nw, metavalue in nwms:  # hier wordt overschreven als er meerdere resultaten zouden zijn
                if nw == correctword:
                    result = (wrongword, correctword, 'Morphology', 'InflectionError', metavalue, bpl_word)
        else:
            result = (wrongword, correctword, 'Lexicology', 'Lexical Error', 'Paraphasy',
                      bpl_word)
    return result


#: dttp = default token tree predicate
def dtp(token, tree): return True


def welnietttp(token: Token, stree: SynTree) -> bool:
    """
    The function *welniettp* checks whether *token* has been analysed as a verb in *stree*
    :param token: input token
    :param stree: input syntactic structure
    :return: True if *token* has been analysed as a verb in *stree*, False otherwise.
    """
    beginval = str(token.pos)
    wordnode = find1(stree, f'.//node[@pt="ww" and @begin="{beginval}"]')
    result = wordnode is not None
    return result


# keer removed
#: The constant *disambiguation_replacements* contains a list of tuples. Each tuple
#: contains  the list of words each of which should be replaced to avoid
#: ambiguity, and their replacement.
#:
#: Currently,
#:
#: * singular neuter nouns that can also be a verb form (e.g. *werk*) are replaced by the word *gas*, which only is a noun;
#: * singular uter nouns that can also be a verb form (e.g. *puzzel*) are replaced by the word *teil*, which is only a noun;
#: * plural nouns that can also be a verb (e.g. *planten*) are replaced by the word *teilen*, which is only a noun;
#: * adjectives without an *-e* ending that can also be a verb (e.g. *dicht*) are replaced by the word *mooi*, which is only an adjective;
#: * adjectives with an *-e* ending that can also be a verb (*e.g. witte*) are replaced by the word *mooie*, which is only an adjective.
#: * the words *wel* and *niet* are replaced by a nonambiguous adverb (*ietsjes*) if they are parsed as a verb in the original parse
#:
#: **Remark** Currently no distinction is made between singular *count* and singular *mass* nouns: they are both replaced by the same word. This may have to be adapted.
disambiguation_replacements: List[Tuple[TokenTreePredicate, List[str], str]] = \
    [(dtp, ['huis', 'water', 'paard', 'werk', 'stuur', 'feest', 'snoep', 'geluid',
                                      'kwartet', 'kruis'], 'gas'),
     (dtp, ['toren', 'fiets', 'puzzel', 'boom', 'vis', 'melk', 'zon', 'pot', 'klok',
            'school', 'boer', 'lepel', 'jas', 'tuin', 'fles', 'lucht', 'emmer', 'maan', 'kachel',
            'kwak', 'verf', 'hop', 'kam', 'spiegel', 'klap', 'stal', 'lijm', 'lift', 'kat',
            'wagen', 'schep', 'kus', 'wind', 'borstel', 'duim', 'strik', 'klik', 'pleister',
            'stok', 'knoop', 'rits', 'knikker', 'put', 'boor', 'plons', 'beurt', 'toeter', 'poot',
            'punt', 'post', 'gom', 'tap', 'kraanwagen', 'drup', 'wieg', 'kriebel', 'pit', 'zaag',
            'slof', 'deuk', 'hark', 'jeuk', 'stift', 'aard', 'hamster', 'kiek', 'haak', 'schroef',
            'tape', 'vorm', 'klem', 'mot', 'druppel'], 'teil'),
     (dtp, ['bomen', 'kussen', 'kaarten', 'beesten', 'weken', 'huizen', 'apen', 'poten',
            'wieken', 'paarden', 'stoelen', 'ramen', 'strepen', 'planten', 'groeten',
            'flessen', 'boeren', 'punten', 'tranen'], 'teilen'),
     (dtp, ['snel', 'wit', 'kort', 'dicht'], 'mooi'),
     (dtp, ['witte'], 'mooie'),
     (welnietttp, ['wel', 'niet'], 'ietsjes')  # find a different adverb that does not get inside constituents (ietsjes?)
     ]


def getdisambiguationdict() -> Dict[str, Tuple[TokenTreePredicate, str]]:
    '''
    :return: a dictionary with words as key and a tuple of a condition and a replacement as values

    The function *getdisambiguationdict* creates a dictionary with word:(cond, replacement)
    items. It selects its content from the constant *disambiguation_replacements*:

    .. autodata:: basicreplacements::disambiguation_replacements
         :no-value:
    '''
    disambiguationdict = {}
    for cond, ws, repl in disambiguation_replacements:
        for w in ws:
            disambiguationdict[w] = cond, repl
    return disambiguationdict



from sastadev import xlsx
from sastadev.conf import settings

gcadd = '_count_tsv'
txtext = ".txt"
tab = '\t'
IDheaders = ['id']
uttcountheaders = ['uttcount', 'uttcounts']


# def oldget_goldcounts(infilename, treebankname):
#     '''
#     Reads the file with name infilename in SASTA GoldCount Format (SGCF)
#     Excel file with one column with header ID,
#     :param infilename:
#     :param treebankname: the name of the file containing the treebank
#     :return: a dictionary  with as  key a tuple (level, item) and as value a Counter  with key uttid and value its count
#     '''
#
#     (_, tail) = os.path.split(treebankname)
#     (cleantreebankname, ext) = os.path.splitext(tail)
#
#     thedata = {}
#
#     # To open Workbook
#     wb = xlrd.open_workbook(infilename)
#     sheet = wb.sheet_by_index(0)
#
#     startrow = 0
#     startcol = 0
#     headerrow = 0
#     lastrow = sheet.nrows
#     lastcol = sheet.ncols
#     idcol = None
#     treebankcol = None
#
#     for rowctr in range(startrow, lastrow):
#         if rowctr == headerrow:
#             for colctr in range(startcol, lastcol):
#                 curcell = sheet.cell_value(rowctr, colctr)
#                 lccurcell = curcell.lower().strip()
#                 if lccurcell in IDheaders:
#                     idcol = colctr
#                 elif curcell == cleantreebankname:
#                     treebankcol = colctr
#         else:
#             if idcol is None:
#                 settings.LOGGER.error('no ID column found')
#                 exit(-1)
#             if treebankcol is None:
#                 settings.LOGGER.error('No column found for Treebank {}.'.format(treebankname))
#                 exit(-1)
#             theid = sheet.cell_value(rowctr, idcol)
#             thecount = sheet.cell_value(rowctr, treebankcol)
#             if thecount != '':
#                 # make it robust here against nonnumber strings
#                 thecountint = int(thecount)
#                 thedata[theid] = thecountint
#
#     return thedata
#
#
# def old2get_goldcounts(infilename):
#     '''
#     Reads the file with name infilename in SASTA GoldCount Format (SGCF)
#     Excel file with one column with header ID, and one header uttcount. All other columns are ignored
#     :param infilename:
#     :return: a dictionary  with as  key a tuple (level, item) and as value a Counter  with key uttid and value its count
#     '''
#
#     thedata = {}
#
#     # To open Workbook
#     wb = xlrd.open_workbook(infilename)
#     sheet = wb.sheet_by_index(0)
#
#     startrow = 0
#     startcol = 0
#     headerrow = 0
#     lastrow = sheet.nrows
#     lastcol = sheet.ncols
#     idcol = None
#     uttcountcol = None
#
#     for rowctr in range(startrow, lastrow):
#         if rowctr == headerrow:
#             for colctr in range(startcol, lastcol):
#                 curcell = sheet.cell_value(rowctr, colctr)
#                 lccurcell = curcell.lower().strip()
#                 if lccurcell in IDheaders:
#                     idcol = colctr
#                 elif lccurcell in uttcountheaders:
#                     uttcountcol = colctr
#         else:
#             if idcol is None:
#                 settings.LOGGER.error('no ID column found')
#                 break
#             elif uttcountcol is None:
#                 settings.LOGGER.error('No uttcount column found')
#                 break
#             else:
#                 theid = sheet.cell_value(rowctr, idcol)
#                 thecount = sheet.cell_value(rowctr, uttcountcol)
#                 if thecount != '':
#                     # make it robust here against nonnumber strings
#                     thecountint = int(thecount)
#                     thedata[theid] = thecountint
#
#     return thedata


def get_goldcounts(infilename):
    '''
    Reads the file with name infilename in SASTA GoldCount Format (SGCF)
    Excel file with one column with header ID, and one header uttcount. All other columns are ignored. Empty string in uttcount is interpreted as 0
    :param infilename:
    :return: a dictionary  with as  key a tuple (level, item) and as value a Counter  with key uttid and value its count
    '''

    thedata = {}
    header, data = xlsx.getxlsxdata(infilename)
    idcol = None

    for col, val in enumerate(header):
        if val.lower() in IDheaders:
            idcol = col
        elif val.lower() in uttcountheaders:
            uttcountcol = col
    if idcol is None:
        settings.LOGGER.error('no ID column found')
        return thedata
    if uttcountcol is None:
        settings.LOGGER.error('No uttcount column found')
        return thedata
    for row in data:
        if len(row) >= max(idcol, uttcountcol):
            theid = row[idcol]
            thecount = row[uttcountcol]
            theintcount = 0 if thecount == '' else int(thecount)
            thedata[theid] = theintcount
        else:
            rowstrlist = [str(x) for x in row]
            rowstr = ';'.join(rowstrlist)
            settings.LOGGER.error(f'Row has no id or uttcount: {rowstr}')
    return thedata



'''
The module *compounds*:

* initialisez the compound dictionary *compounds*, which is a multidimensional Python dictionary
  Dict[str, Dict[int, str]], which maps a string (for a lemma in CELEX orthography HeadDiaNew)
  and a column number to the value of the cell with this column number in the CSV file from which it is derived:

   * e.g., compounds["aalbes"][0] == "NN"
   * e.g., compounds["aalbes"][2] == "N"


 The file from which it is derived is the CSV file (with backslash as separator) *Ncompounds-attempt2.txt*
 in subfolder compoundfiles in the code folder. It contains a header and three fields in each row:

  * FlatClass: the part of speech codes for the parts (lexical parts and affixes) e.g. NN for *aalbes*
  * HeadDiaNew: the orthography of the lemma, e.g. *aalbes*
  * Class: the part of speech code for the whole word, e.g. N for *aalbes*

  The file has been derived from the CELEX database.

* provides the functions getcompounds and iscompounds:
   * .. autofunction:: getcompounds
   * .. autofunction:: iscompound


'''


import csv
import os
from collections import defaultdict
from typing import Dict, List

from sastadev.conf import settings
from sastadev.sastatypes import SynTree
from sastadev.treebankfunctions import getattval

underscore = "_"
FlatClass = 0
HeadDiaNew = 1
Class = 2

Headers = {}
Headers[FlatClass] = "FlatClass"
Headers[HeadDiaNew] = "HeadDiaNew"
Headers[Class] = "Class"

comma = ","

dictfilename = os.path.join(settings.SD_DIR, 'data', 'compoundfiles', 'Ncompounds-attempt2.txt')
dictfile = open(dictfilename, 'r', encoding='utf8')

getwordsxpath = ".//node[@pt]"


def getcompounds(syntree: SynTree) -> List[SynTree]:
    '''
    .. _getcompounds-label:

    The function getcompounds takes a syntactic structure as input and returns a list of syntactic structures
    (nodes for words) that are compounds. It implements TARSP language measure T086 (SamZn). A node is considered
    to be a compound if its *pt* attribute has the value *n* (is a noun) and its lemma meets the
    condition iscompound(lemma).

    '''
    results = []
    tlist = syntree.xpath(getwordsxpath)
    for t in tlist:
        w = getattval(t, 'word')
        lemma = getattval(t, 'lemma')
        pt = getattval(t, 'pt')
        if pt == 'n' and iscompound(lemma):
            results.append(t)
    return results

# I do not know how to type this, because the nesting can be arbitrarily deep


def nested_dict(n, type):
    if n == 1:
        return defaultdict(type)
    else:
        return defaultdict(lambda: nested_dict(n - 1, type))


def iscompound(str: str) -> bool:
    '''
    The function iscompound checks whether the input string str is a compound.
    That is the case if

    * either it contains an underscore (lemmas of  strings recognised as a compound by Alpino contain an underscore),
    * or it is contained in the compound dictionary *compounds*.


    '''
    if underscore in str:
        result = True
    else:
        result = str in compounds
    return result


mysep = "\\"
myquotechar = ''

compounds: Dict[str, Dict[int, str]] = nested_dict(2, str)


# there is an error here, the file has a header, and that is also read in as a compound
# we should make  the character encoding explicit
settings.LOGGER.info("Initializing compound module...")
myreader = csv.reader(dictfile, delimiter=mysep)
for row in myreader:
    compounds[row[HeadDiaNew]][FlatClass] = row[FlatClass]
    compounds[row[HeadDiaNew]][Class] = row[Class]



from lxml import etree

from sastadev.macros import expandmacros

expectedresultslist = [[],
                       [('top', '0', '8'), ('du', '0', '7'),
                        ('pp', '4', '7'), ('np', '5', '7')],
                       [('top', '0', '8'), ('du', '0', '7'),
                        ('pp', '4', '7'), ('np', '5', '7')],
                       [('smain', '1', '7')],
                       [('smain', '1', '7')],
                       [],
                       [('smain', '1', '7')],
                       [], [], [], [], [], [], [], [], [], [], []]

bigquery = """//node[(

 %declarative%

 and

 %Ond%

 and

 %Tarsp_B_X_count% = 3

 )]
"""


def getresults(stree, fullquery):
    noderesults = stree.xpath(fullquery)
    results = []
    for noderesult in noderesults:
        cat = noderesult.get('cat')
        pos = noderesult.get('pt')
        poscat = pos if cat is None else cat
        begin = noderesult.get('begin')
        end = noderesult.get('end')
        result = (poscat, begin, end)
        results.append(result)
    return results


def main():
    queries = []
    queries.append('//node[%Tarsp_WVz_exception%]')
    queries.append('//node[count(node[@cat or @pt]) = 2]')
    queries.append('//node[count(node[@cat] |  node[@pt]) = 2]')
    queries.append('//node[( %declarative% )]')
    queries.append('//node[( %Ond%  )]')
    queries.append('//node[(   %Tarsp_B_X_count% = 3)]')
    # queries.append(bigquery)
    queries.append('//node[%declarative% and %Ond%]')
    queries.append('//node[%declarative% and %Tarsp_B_X_count% = 3]')
    queries.append('//node[%declarative% and (%Tarsp_B_X_count%) = 3]')
    queries.append('//node[%declarative% and (%Tarsp_B_X_count% = 3)]')
    queries.append('//node[%declarative% and ((%Tarsp_B_X_count%) = 3)]')
    queries.append('//node[%Ond% and %Tarsp_B_X_count% = 3]')
    queries.append('//node[%Ond% and (%Tarsp_B_X_count%) = 3]')
    queries.append('//node[%Ond% and (%Tarsp_B_X_count% = 3)]')
    queries.append(
        '//node[( (%declarative%) and (%Ond%)  and  (%Tarsp_B_X_count% = 3))]')
    queries.append(
        '//node[( (%declarative%) and (%Ond%)  and  (%Tarsp_B_X_count% = 3))]')
    queries.append(
        '//node[( %declarative% and %Ond%  and  (%Tarsp_B_X_count%) = 3)]')
    queries.append(
        '//node[(%declarative% and %Ond% and %Tarsp_W% and %Tarsp_B_X_count% = 3 and %realcomplormodnodecount% = 4)]')

    fullqueries = []
    for query in queries:
        expandedquery = expandmacros(query)
        fullqueries.append('.' + expandedquery)

    streestrings = {}
    streestrings[0] = """
    <alpino_ds version="1.6">
      <parser cats="1" skips="0" />
      <node begin="0" cat="top" end="8" id="0" rel="top">
        <node begin="0" cat="du" end="7" id="1" rel="--">
          <node begin="0" conjtype="neven" end="1" frame="complementizer(root)" his="normal" his_1="normal" id="2" lcat="du" lemma="en" pos="comp" postag="VG(neven)" pt="vg" rel="dlink" root="en" sc="root" sense="en" word="en"/>
          <node begin="1" cat="smain" end="7" id="3" rel="nucl">
            <node begin="1" end="2" frame="tmp_adverb" his="normal" his_1="normal" id="4" lcat="advp" lemma="toen" pos="adv" postag="BW()" pt="bw" rel="mod" root="toen" sense="toen" special="tmp" word="toen"/>
            <node begin="2" end="3" frame="verb(hebben,past(sg),ld_pp)" his="normal" his_1="normal" id="5" infl="sg" lcat="smain" lemma="zitten" pos="verb" postag="WW(pv,verl,ev)" pt="ww" pvagr="ev" pvtijd="verl" rel="hd" root="zit" sc="ld_pp" sense="zit" stype="declarative" tense="past" word="zat" wvorm="pv"/>
            <node begin="3" case="nom" def="def" end="4" frame="pronoun(nwh,thi,sg,de,nom,def)" gen="de" genus="masc" getal="ev" his="normal" his_1="normal" id="6" lcat="np" lemma="hij" naamval="nomin" num="sg" pdtype="pron" per="thi" persoon="3" pos="pron" postag="VNW(pers,pron,nomin,vol,3,ev,masc)" pt="vnw" rel="su" rnum="sg" root="hij" sense="hij" status="vol" vwtype="pers" wh="nwh" word="hij"/>
            <node begin="4" cat="pp" end="7" id="7" rel="ld">
              <node begin="4" end="5" frame="preposition(in,[])" his="normal" his_1="normal" id="8" lcat="pp" lemma="in" pos="prep" postag="VZ(init)" pt="vz" rel="hd" root="in" sense="in" vztype="init" word="in"/>
              <node begin="5" cat="np" end="7" id="9" rel="obj1">
                <node begin="5" end="6" frame="determiner(de)" his="normal" his_1="normal" id="10" infl="de" lcat="detp" lemma="de" lwtype="bep" naamval="stan" npagr="rest" pos="det" postag="LID(bep,stan,rest)" pt="lid" rel="det" root="de" sense="de" word="de"/>
                <node begin="6" end="7" frame="noun(both,both,both)" gen="both" genus="zijd" getal="ev" graad="basis" his="noun" id="11" lcat="np" lemma="weiland" naamval="stan" ntype="soort" num="both" pos="noun" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="hd" rnum="sg" root="weiland" sense="weiland" word="weiland"/>
              </node>
            </node>
          </node>
        </node>
        <node begin="7" end="8" frame="punct(punt)" his="normal" his_1="normal" id="12" lcat="punct" lemma="." pos="punct" postag="LET()" pt="let" rel="--" root="." sense="." special="punt" word="."/>
      </node>
      <sentence sentid="42">en toen zat hij in de weiland .</sentence>
    <metadata>
    <meta type="text" name="charencoding" value="UTF8" />
    <meta type="text" name="childage" value="9;6" />
    <meta type="int" name="childmonths" value="114" />
    <meta type="text" name="comment" value="##META text title = STAP_04" />
    <meta type="text" name="session" value="STAP_04" />
    <meta type="text" name="origutt" value="en toen zat hij in de weiland." />
    <meta type="text" name="parsefile" value="Unknown_corpus_STAP_04_u00000000042.xml" />
    <meta type="text" name="speaker" value="CHI" />
    <meta type="int" name="uttendlineno" value="90" />
    <meta type="int" name="uttid" value="42" />
    <meta type="int" name="uttstartlineno" value="90" />
    <meta type="text" name="name" value="chi" />
    <meta type="text" name="SES" value="" />
    <meta type="text" name="age" value="9;6" />
    <meta type="text" name="custom" value="" />
    <meta type="text" name="education" value="" />
    <meta type="text" name="group" value="" />
    <meta type="text" name="language" value="nld" />
    <meta type="int" name="months" value="114" />
    <meta type="text" name="role" value="Target_Child" />
    <meta type="text" name="sex" value="male" />
    <meta type="text" name="xsid" value="42" />
    <meta type="int" name="uttno" value="42" />
    </metadata>
    </alpino_ds>"""

    strees = {}
    for el in streestrings:
        strees[el] = etree.fromstring(streestrings[el])

    allresults = []
    for stree in strees:
        for query, fullquery, expectedresults in zip(queries, fullqueries,
                                                     expectedresultslist):
            # print(query)
            results = getresults(strees[stree], fullquery)
            assert results == expectedresults
            # allresults.append(results)

    # print(allresults)
    # print('finished')


if __name__ == '__main__':
    main()



from lxml import etree

from sastadev.treebankfunctions import (adjacent, find1, getnodeyield,
                                        immediately_follows,
                                        immediately_precedes)

streestrings = {}
streestrings[1] = '''
<alpino_ds version="1.6">
  <parser cats="1" skips="5" />
  <node begin="0" cat="top" end="8" id="0" rel="top">
    <node begin="0" conjtype="neven" end="1" frame="complementizer(root)" his="robust_skip" id="1" lcat="--" lemma="en" pos="comp" postag="VG(neven)" pt="vg" rel="--" root="en" sc="root" sense="en" word="en"/>
    <node begin="1" end="2" frame="--" genus="zijd" getal="ev" graad="basis" his="robust_skip" id="2" lcat="--" lemma="uhm" naamval="stan" ntype="soort" pos="--" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="--" root="uhm" sense="uhm" word="uhm"/>
    <node begin="2" conjtype="neven" end="3" frame="conj(en)" his="robust_skip" id="3" lcat="--" lemma="en" pos="vg" postag="VG(neven)" pt="vg" rel="--" root="en" sense="en" word="en"/>
    <node begin="3" end="4" frame="--" genus="zijd" getal="ev" graad="basis" his="robust_skip" id="4" lcat="--" lemma="uhm" naamval="stan" ntype="soort" pos="--" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="--" root="uhm" sense="uhm" word="uhm"/>
    <node begin="4" case="nom" def="def" end="5" frame="pronoun(nwh,thi,sg,de,nom,def)" gen="de" genus="masc" getal="ev" his="robust_skip" id="5" lcat="--" lemma="hij" naamval="nomin" num="sg" pdtype="pron" per="thi" persoon="3" pos="pron" postag="VNW(pers,pron,nomin,vol,3,ev,masc)" pt="vnw" rel="--" root="hij" sense="hij" status="vol" vwtype="pers" wh="nwh" word="hij"/>
    <node begin="5" cat="smain" end="8" id="6" rel="--">
      <node begin="5" case="nom" def="def" end="6" frame="pronoun(nwh,thi,sg,de,nom,def)" gen="de" genus="masc" getal="ev" his="normal" his_1="normal" id="7" lcat="np" lemma="hij" naamval="nomin" num="sg" pdtype="pron" per="thi" persoon="3" pos="pron" postag="VNW(pers,pron,nomin,vol,3,ev,masc)" pt="vnw" rel="su" rnum="sg" root="hij" sense="hij" status="vol" vwtype="pers" wh="nwh" word="hij"/>
      <node begin="6" end="7" frame="verb(unacc,sg_heeft,intransitive)" his="normal" his_1="normal" id="8" infl="sg_heeft" lcat="smain" lemma="zijn" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="ben" sc="intransitive" sense="ben" stype="declarative" tense="present" word="is" wvorm="pv"/>
      <node begin="7" end="8" frame="adverb" his="normal" his_1="normal" id="9" lcat="advp" lemma="nogal" pos="adv" postag="BW()" pt="bw" rel="mod" root="nogal" sense="nogal" word="nogal"/>
    </node>
  </node>
  <sentence sentid="32">en uhm en uhm hij hij is nogal</sentence>
<metadata>
<meta type="text" name="charencoding" value="UTF8" />
<meta type="text" name="childage" value="" />
<meta type="text" name="childmonths" value="" />
<meta type="text" name="comment" value="##META text samplenaam = ASTA-06" />
<meta type="text" name="session" value="ASTA_sample_06" />
<meta type="text" name="origutt" value="en uhm en uhm hij hij is nogal " />
<meta type="text" name="parsefile" value="Unknown_corpus_ASTA_sample_06_u00000000046.xml" />
<meta type="text" name="speaker" value="PMA" />
<meta type="int" name="uttendlineno" value="85" />
<meta type="int" name="uttid" value="32" />
<meta type="int" name="uttstartlineno" value="85" />
<meta type="text" name="name" value="pma" />
<meta type="text" name="SES" value="" />
<meta type="text" name="age" value="" />
<meta type="text" name="custom" value="" />
<meta type="text" name="education" value="" />
<meta type="text" name="group" value="" />
<meta type="text" name="language" value="nld" />
<meta type="text" name="months" value="" />
<meta type="text" name="role" value="Other" />
<meta type="text" name="sex" value="" />
<meta type="text" name="xsid" value="32" />
<meta type="int" name="uttno" value="46" />
</metadata>
</alpino_ds>
'''

streestrings[2] = '''
<alpino_ds version="1.6">
  <parser cats="3" skips="0" />
  <node begin="0" cat="top" end="17" id="0" rel="top">
    <node begin="0" cat="du" end="16" id="1" rel="--">
      <node begin="0" cat="smain" end="3" id="2" rel="dp">
        <node begin="0" case="nom" def="def" end="1" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" his="normal" his_1="normal" id="3" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik"/>
        <node begin="1" end="2" frame="verb(hebben,sg1,transitive_ndev)" his="normal" his_1="normal" id="4" infl="sg1" lcat="smain" lemma="hebben" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="heb" sc="transitive_ndev" sense="heb" stype="declarative" tense="present" word="heb" wvorm="pv"/>
        <node begin="2" case="both" def="indef" end="3" frame="pronoun(nwh,thi,sg,both,both,indef,strpro)" gen="both" his="normal" his_1="normal" id="5" lcat="np" lemma="één" num="sg" numtype="hoofd" per="thi" pos="pron" positie="vrij" postag="TW(hoofd,vrij)" pt="tw" rel="obj1" rnum="sg" root="één" sense="één" special="strpro" wh="nwh" word="een"/>
      </node>
      <node begin="3" cat="smain" end="6" id="6" rel="dp">
        <node begin="3" case="nom" def="def" end="4" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" his="normal" his_1="normal" id="7" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik"/>
        <node begin="4" end="5" frame="verb(hebben,sg1,transitive_ndev)" his="normal" his_1="normal" id="8" infl="sg1" lcat="smain" lemma="hebben" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="heb" sc="transitive_ndev" sense="heb" stype="declarative" tense="present" word="heb" wvorm="pv"/>
        <node begin="5" case="both" def="indef" end="6" frame="pronoun(nwh,thi,sg,both,both,indef,strpro)" gen="both" his="normal" his_1="normal" id="9" lcat="np" lemma="één" num="sg" numtype="hoofd" per="thi" pos="pron" positie="vrij" postag="TW(hoofd,vrij)" pt="tw" rel="obj1" rnum="sg" root="één" sense="één" special="strpro" wh="nwh" word="een"/>
      </node>
      <node begin="6" cat="smain" end="16" id="10" rel="dp">
        <node begin="6" case="nom" def="def" end="7" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" his="normal" his_1="normal" id="11" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik"/>
        <node begin="7" end="8" frame="verb(hebben,sg1,transitive_ndev)" his="normal" his_1="normal" id="12" infl="sg1" lcat="smain" lemma="hebben" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="heb" sc="transitive_ndev" sense="heb" stype="declarative" tense="present" word="heb" wvorm="pv"/>
        <node begin="8" cat="np" end="16" id="13" rel="obj1">
          <node begin="8" end="9" frame="determiner(een)" his="normal" his_1="normal" id="14" infl="een" lcat="detp" lemma="een" lwtype="onbep" naamval="stan" npagr="agr" pos="det" postag="LID(onbep,stan,agr)" pt="lid" rel="det" root="een" sense="een" word="een"/>
          <node begin="9" end="10" frame="noun(de,count,bare_meas)" gen="de" genus="zijd" getal="ev" graad="basis" his="normal" his_1="normal" id="15" lcat="np" lemma="man" naamval="stan" ntype="soort" num="bare_meas" pos="noun" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="hd" rnum="sg" root="man" sense="man" word="man"/>
          <node begin="10" cat="rel" end="16" id="16" rel="mod">
            <node begin="10" cat="pp" end="12" id="17" index="1" rel="rhd">
              <node begin="10" end="11" frame="preposition(met,[mee,[en,al]])" his="normal" his_1="normal" id="18" lcat="pp" lemma="met" pos="prep" postag="VZ(init)" pt="vz" rel="hd" root="met" sense="met" vztype="init" word="met"/>
              <node begin="11" case="obl" end="12" frame="rel_pronoun(both,obl)" gen="both" getal="getal" his="normal" his_1="normal" id="19" lcat="np" lemma="wie" naamval="stan" pdtype="pron" persoon="3p" pos="pron" postag="VNW(vb,pron,stan,vol,3p,getal)" pt="vnw" rel="obj1" rnum="sg" root="wie" sense="wie" status="vol" vwtype="vb" wh="rel" word="wie"/>
            </node>
            <node begin="10" cat="ssub" end="16" id="20" rel="body">
              <node begin="12" case="nom" def="def" end="13" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" his="normal" his_1="normal" id="21" index="2" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik"/>
              <node begin="13" end="14" frame="verb(hebben,modal_not_u,modifier(aux(inf)))" his="normal" his_1="normal" id="22" infl="modal_not_u" lcat="ssub" lemma="willen" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="wil" sc="modifier(aux(inf))" sense="wil" tense="present" word="wil" wvorm="pv"/>
              <node begin="10" cat="inf" end="16" id="23" rel="vc">
                <node begin="12" end="13" id="24" index="2" rel="su"/>
                <node begin="14" buiging="zonder" end="15" frame="verb(zijn,inf(no_e),aux(inf))" his="normal" his_1="normal" id="25" infl="inf(no_e)" lcat="inf" lemma="gaan" pos="verb" positie="vrij" postag="WW(inf,vrij,zonder)" pt="ww" rel="hd" root="ga" sc="aux(inf)" sense="ga" word="gaan" wvorm="inf"/>
                <node begin="10" cat="inf" end="16" id="26" rel="vc">
                  <node begin="10" end="12" id="27" index="1" rel="pc"/>
                  <node begin="12" end="13" id="28" index="2" rel="su"/>
                  <node begin="15" buiging="zonder" end="16" frame="verb(zijn,inf,pc_pp(met))" his="normal" his_1="normal" id="29" infl="inf" lcat="inf" lemma="trouwen" pos="verb" positie="vrij" postag="WW(inf,vrij,zonder)" pt="ww" rel="hd" root="trouw" sc="pc_pp(met)" sense="trouw-met" word="trouwen" wvorm="inf"/>
                </node>
              </node>
            </node>
          </node>
        </node>
      </node>
    </node>
    <node begin="16" end="17" frame="--" genus="zijd" getal="ev" graad="basis" his="skip" id="30" lcat="--" lemma="uhm" naamval="stan" ntype="soort" pos="--" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="--" root="uhm" sense="uhm" word="uhm"/>
  </node>
  <sentence sentid="29">ik heb een ik heb een ik heb een man met wie ik wil gaan trouwen uhm</sentence>
<metadata>
<meta type="text" name="charencoding" value="UTF8" />
<meta type="text" name="childage" value="" />
<meta type="text" name="childmonths" value="" />
<meta type="text" name="comment" value="##META text samplenaam = ASTA-06" />
<meta type="text" name="session" value="ASTA_sample_06" />
<meta type="text" name="origutt" value="ik heb een ik heb een ik heb een man met wie ik wil gaan trouwen uhm " />
<meta type="text" name="parsefile" value="Unknown_corpus_ASTA_sample_06_u00000000042.xml" />
<meta type="text" name="speaker" value="PMA" />
<meta type="int" name="uttendlineno" value="78" />
<meta type="int" name="uttid" value="29" />
<meta type="int" name="uttstartlineno" value="78" />
<meta type="text" name="name" value="pma" />
<meta type="text" name="SES" value="" />
<meta type="text" name="age" value="" />
<meta type="text" name="custom" value="" />
<meta type="text" name="education" value="" />
<meta type="text" name="group" value="" />
<meta type="text" name="language" value="nld" />
<meta type="text" name="months" value="" />
<meta type="text" name="role" value="Other" />
<meta type="text" name="sex" value="" />
<meta type="text" name="xsid" value="29" />
<meta type="int" name="uttno" value="42" />
</metadata>
</alpino_ds>
'''

strees = {}
for el in streestrings:
    strees[el] = etree.fromstring(streestrings[el])


def getnode(stree, word):
    result = find1(stree, f'//node[@word="{word}"]')
    return result

# is nogal


thetree = strees[1]
node1 = getnode(thetree, 'is')
node2 = getnode(thetree, 'nogal')
result = adjacent(node1, node2, thetree)
assert result is True

thetree = strees[2]
node1 = getnode(thetree, 'gaan')
node2 = getnode(thetree, 'trouwen')
result = adjacent(node1, node2, thetree)
assert result is True

thetree = strees[2]
node1 = getnode(thetree, 'wil')
node2 = getnode(thetree, 'gaan')
result = adjacent(node1, node2, thetree)
assert result is True

thetree = strees[2]
node1 = getnode(thetree, 'wil')
node2 = getnode(thetree, 'trouwen')
result = adjacent(node1, node2, thetree)
assert result is False

thetree = strees[2]
yieldnodes = getnodeyield(thetree)
for i, n in enumerate(yieldnodes):
    if i > 0:
        assert adjacent(yieldnodes[i - 1], n, thetree)
        assert immediately_precedes(yieldnodes[i - 1], n, thetree)
        assert immediately_follows(n, yieldnodes[i - 1], thetree)
    if i < len(yieldnodes) - 1:
        assert adjacent(n, yieldnodes[i + 1], thetree)
        assert immediately_precedes(n, yieldnodes[i + 1], thetree)
        assert immediately_follows(yieldnodes[i + 1], n, thetree)



'''
This module contains definitions of types used in multiple modules
'''

from collections import Counter
from typing import Callable, Dict, List, Optional, Tuple, Union

from lxml import etree  # type: ignore
from typing_extensions import TypeAlias

from sastadev.query import Query
from sastadev.sastatoken import Token

AltId: TypeAlias = int
BackPlacement: TypeAlias = int
CapitalName: TypeAlias = str
CountryName: TypeAlias = str
ContinentName: TypeAlias = str
CELEXPosCode: TypeAlias = str
FirstName: TypeAlias = str
LocationName: TypeAlias = str
DCOIPt: TypeAlias = str
DeHet: TypeAlias = str
CELEX_INFL: TypeAlias = str
DCOITuple: TypeAlias = Tuple
Lemma: TypeAlias = str
CorrectionMode: TypeAlias = str  # Literal['0','1','n']
ErrorDict: TypeAlias = Dict[str, List[List[str]]]
Level: TypeAlias = str  # in the future perhaps NewType('Level', str)
Item: TypeAlias = str  # in the future perhaps NewType('Item', str)
Item_Level: TypeAlias = Tuple[Item, Level]
IntSpan: TypeAlias = Tuple[int, int]
AltCodeDict: TypeAlias = Dict[Item_Level, Item_Level]
QId: TypeAlias = str  # in the futute perhaps NewType('QId', str)
UttId: TypeAlias = str  # in the future perhaps NewType('UttId', str)
Position: TypeAlias = int  # in the future perhapos NewType('Position', int)
PositionStr: TypeAlias = str
Stage: TypeAlias = int
SynTree: TypeAlias = etree._Element  # type: ignore
GoldTuple: TypeAlias = Tuple[str, str, Counter]
Item2LevelsDict: TypeAlias = Dict[Item, List[Level]]
Match: TypeAlias = Tuple[SynTree, SynTree]
Matches: TypeAlias = List[Match]
MatchesDict: TypeAlias = Dict[Tuple[QId, UttId], Matches]
MetaElement: TypeAlias = etree.Element
ExactResult: TypeAlias = Tuple[UttId, Position]
ExactResults: TypeAlias = List[ExactResult]
ExactResultsDict: TypeAlias = Dict[QId, ExactResults]  # qid
Gender: TypeAlias = str
Penalty: TypeAlias = int
PhiTriple: TypeAlias = Tuple[str, str, str]
OptPhiTriple: TypeAlias = Optional[PhiTriple]
PositionMap: TypeAlias = Dict[Position, Position]
QueryDict: TypeAlias = Dict[QId, Query]
QIdCount: TypeAlias = Dict[QId, int]
MethodName: TypeAlias = str  # perhaps in the future NewType('MethodName', str)
FileName: TypeAlias = str  # perhaps in the future NewType('FileName', str)
ReplacementMode: TypeAlias = int
ResultsCounter: TypeAlias = Counter  # Counter[UttId]  # Dict[UttId, int]
ResultsDict: TypeAlias = Dict[QId, ResultsCounter]
Span: TypeAlias = Tuple[PositionStr, PositionStr]
Item_Level2QIdDict: TypeAlias = Dict[Item_Level, QId]
Nort: TypeAlias = Union[SynTree, Token]
ExactResultsFilter: TypeAlias = Callable[[Query, ExactResultsDict, ExactResult], bool]
Targets: TypeAlias = int
Treebank: TypeAlias = etree.Element
TreePredicate: TypeAlias = Callable[[SynTree], bool]
TokenTreePredicate: TypeAlias = Callable[[Token, SynTree], bool]
URL: TypeAlias = str
UttTokenDict: TypeAlias = Dict[UttId, List[Token]]
UttWordDict: TypeAlias = Dict[UttId, List[str]]
WordInfo: TypeAlias = Tuple[Optional[CELEXPosCode], Optional[DeHet], Optional[CELEX_INFL], Optional[Lemma]]
# moved the following to allresuls.py
#CoreQueryFunction: TypeAlias = Callable[[SynTree], List[SynTree]]
#PostQueryFunction: TypeAlias = Callable[[SynTree, allresults.AllResults], List[SynTree]]
#QueryFunction: TypeAlias = Union[CoreQueryFunction, PostQueryFunction]



''''
This module is not used anymore
'''
from math import log

knownmethods = {'asta', 'stap', 'tarsp'}


def isknownmethod(mname):
    if mname is None:
        result = False
    else:
        result = mname.upper() in knownmethods
    return result


def getfnbases(method, maxval):
    model = knownmethods[method].basenamemodel
    width = max([2, int(log(maxval, 10)) + 1])
    results = []
    for i in range(1, maxval + 1):
        istr = str(i).rjust(width, '0')
        result = model.format(istr)
        results.append(result)
    return results



'''
TARSP Screening Procedre (determining the stage of grammatical development)
Based on Schlichting 2015 Chapter Screening (p. 94- 96)

Parameters:
    stage1threshold: implements 'exclusively or almost exclusively one word sentences, set to 95%
    uttcountthreshold: minimum # utterances in the sample, set to 200 based on Schlichting 2005:94


function to expose to other modules is: screening4stage(uttcount, results)

'''
from sastadev.allresults import scores2counts
from sastadev.conf import settings

stage1threshold = 95
uttcountthreshold = 200

OndB = 'T064'
VCW = 'T099'
OndVC = 'T071'
OndW = 'T072'
BW = 'T028'
OndWVC = 'T076'
ik = 'T047'
HwwZ = 'T046'
VCWG = 'T101'
WWWG = 'T127'
BWWG = 'T019'
OndWG = 'T070'
OndWBVC = 'T075'
VzBepZn = 'T116'
en = 'T036'
het = 'T041'
VCWOndBB = 'T100'
OndWVCVCoptX = 'T077'
Nevenschikkende = 'T061'
VCbijzin = 'T098'
S6plus = 'T003'
BBijzin = 'T011'
Verbnieten = ''  # @@
wij = 'T128'
Wond5 = 'T131'
Vr5plus = 'T113'
hetZn = 'T042'
ditdatZn = 'T034'
Verbnietenalsmaar = ''  # @@
verltijd = 'T105'
Zn = 'T145'
BvB = 'T023'
W = 'T119'


def OK(queryid, results):
    result = queryid in results and results[queryid] > 0
    return result


def qcount(queryid, results):
    if queryid in results:
        result = results[queryid]
    else:
        result = 0
    return result


def stage2(results):
    cond1 = OK(OndB, results) or OK(OndVC, results)
    cond2 = OK(VCW, results) or OK(OndW, results) or OK(BW, results)
    result = cond1 and cond2
    return result


def stage3(results):
    WGOK = OK(VCWG, results) or OK(WWWG, results) or OK(BWWG, results) or OK(OndWG, results)
    OndWVCOK = OK(OndWVC, results)
    ikOK = OK(ik, results)
    HwwZOK = OK(HwwZ, results)
    baselist = [OndWVCOK, ikOK, HwwZOK, WGOK]
    resultlist = [b for b in baselist if b]
    result = len(resultlist) >= 3
    return result


def stage4(results):
    OndWBVCOK = OK(OndWBVC, results)
    VzBepZnOK = OK(VzBepZn, results)
    enOK = OK(en, results)
    hetOK = OK(het, results)
    baselist = [OndWBVCOK, VzBepZnOK, enOK, hetOK]
    resultlist = [b for b in baselist if b]
    result = len(resultlist) >= 3
    return result


def stage5(results):
    VCWOndBBOK = OK(VCWOndBB, results)
    OndWVCVCoptXOK = OK(OndWVCVCoptX, results)
    NevenschikkendeOK = OK(Nevenschikkende, results)
    VCbijzinOK = OK(VCbijzin, results)
    S6plusOK = OK(S6plus, results)
    BBijzinOK = OK(BBijzin, results)
    VerbnietenOK = OK(Verbnieten, results)
    wijOK = OK(wij, results)
    cond1 = VCWOndBBOK or OndWVCVCoptXOK
    cond2 = NevenschikkendeOK or VCbijzinOK or S6plusOK or BBijzinOK
    baselist = [cond1, cond2, VerbnietenOK, wijOK]
    resultlist = [b for b in baselist if b]
    result = len(resultlist) >= 3
    return result


def stage6(results):
    Wond5OK = OK(Wond5, results)
    Vr5plusOK = OK(Vr5plus, results)
    hetZnOK = OK(hetZn, results)
    ditdatZnOK = OK(ditdatZn, results)
    VerbnietenalsmaarOK = OK(Verbnietenalsmaar, results)
    verltijdOK = OK(verltijd, results)
    cond1 = Wond5OK or Vr5plusOK
    cond2 = hetZnOK or ditdatZnOK
    baselist = [cond1, cond2, VerbnietenalsmaarOK, verltijdOK]
    resultlist = [b for b in baselist if b]
    result = len(resultlist) >= 3
    return result


def stage1(results):
    BvBqcount = qcount(BvB, results)
    Wqcount = qcount(W, results)
    Znqcount = qcount(Zn, results)
    stage1count = BvBqcount + Wqcount + Znqcount
    allresults = [results[qid] for qid in results]
    sumallresults = sum(allresults)
    if sumallresults != 0:
        proportion = stage1count / sumallresults * 100
    else:
        proportion = 0
        settings.LOGGER.warning('No results found. Output unreliable')
    result = proportion >= stage1threshold
    return result


def screening(results):
    stages = {}
    stages[6] = stage6(results)
    stages[5] = stage5(results)
    stages[4] = stage4(results)
    stages[3] = stage3(results)
    stages[2] = stage2(results)
    stages[1] = stage1(results) and not stages[2]
    result = None
    for s in range(6, 0, -1):
        #        stagesLOGGER.info('Checking stage %s', s)
        if result is None and stages[s]:
            result = s
            # for ls in range(s,0,-1):
            #    if not stages[ls]:
            #        print('Warning. Stage is {} but lower stage {} conditions are not met'. format(s,ls), file=sys.stderr)
    return result


def tarsp_screening(allresults, _):
    resultscounts = scores2counts(allresults.coreresults)
    result = screening4stage(allresults.uttcount, resultscounts)
    return result

# def screening4stage(uttcount, results, thefilename):


def screening4stage(uttcount, results):
    '''
    determines the stage of grammaticl development based on two inputs:
    :param uttcount: size of the sample in # utterances. Values lower than 200 lead to unreliable results
    :param results: dictionary containing (queryid, count) pairs obtained for the sample by sasta doqueries
    :return:
    '''
    #   ;param thefilename: name of the treebank file
    #    (inbase, ext) = os.path.splitext(thefilename)
    #    stagesexplainfile = inbase + 'stage_explanation' + '.txt'
    #    stageslogger = logging.getLogger('Stages:')
    #    stagesLOGGER.setLevel(logging.INFO)
    #    ch = logging.FileHandler(stagesexplainfile)
    #    ch.setLevel(logging.INFO)
#    stagesLOGGER.addHandler(ch)

    if uttcount < uttcountthreshold:
        message = 'TARSP Screening: Less than {} utterances ({}). Results not reliable'.format(uttcountthreshold, uttcount)
        settings.LOGGER.warning(message)
        # print(message)
    result = screening(results)
    return result



'''
The module cleanCHILDEStokens provides the function cleantext to tokenize and clean utterances,
yielding a cleaned text and associated metadata.

* ..autofunction:: cleantext

'''
import re
from copy import deepcopy
from typing import List, Optional, Pattern, TextIO, Tuple, Union

from sastadev import CHAT_Annotation, sastatok
from sastadev.conf import settings
from sastadev.metadata import Meta, Metadata, bpl_none
from sastadev.sastatoken import Token, show
from sastadev.sastatypes import IntSpan

hexformat = '\\u{0:04X}'


space = ' '

scope_open = '<'
scope_close = '>'

bstate, mstate, estate = 0, 1, 2

bstate, ostate, oostate, costate, ccstate = 0, 1, 2, 3, 4

#this should be identical to the checkpattern of cleanCHILDESMD
# #checkpattern = re.compile(r'[][\(\)&%@/=><_0^~↓↑↑↓⇗↗→↘⇘∞≈≋≡∙⌈⌉⌊⌋∆∇⁎⁇°◉▁▔☺∬Ϋ123456789·\u22A5\u00B7\u0001\u2260\u21AB]')
# checkpattern = re.compile(r'[][\(\)&%@/=><_0^~↓↑↑↓⇗↗→↘⇘∞≈≋≡∙⌈⌉⌊⌋∆∇⁎⁇°◉▁▔☺∬Ϋ·\u22A5\u00B7\u0001\u2260\u21AB]')
# # + should not occur except as compound marker black+board
# # next one split up in order to do substitutions
# pluspattern = re.compile(r'(\W)\+|\+(\W)')
# pluspattern1 = re.compile(r'(\W)\+')
# pluspattern2 = re.compile(r'\+(\W)')
illegalcleanedchatsymbols = '<>'


def findscopeclose(tokens: List[Token], offset: int = 0) -> Optional[IntSpan]:
    tokenctr = 0
    bracketcounter = -1
    begin = None
    end = None
    for token in tokens:
        if token.word == scope_open:
            if bracketcounter == -1:
                begin = tokenctr
            bracketcounter += 1
        elif token.word == scope_close:
            if bracketcounter == 0:
                end = tokenctr
                break
            else:
                bracketcounter -= 1
        else:
            pass
        tokenctr += 1
    result = (begin + offset, end + offset) if begin is not None and end is not None else None
    return result


def clearnesting(intokens: List[Token], repkeep: bool) -> Tuple[List[Token], Metadata]:
    tokens = deepcopy(intokens)
    bracketcounter = -1
    newtokens = []
    metadata = []
    bopenfound = False
    tokenctr = 0
    ltokens = len(tokens)
    while tokenctr < ltokens:
        token = tokens[tokenctr]
        if token.word == scope_open:
            begin = tokenctr
            span = findscopeclose(tokens[tokenctr:], offset=tokenctr)
            if span is None:
                settings.LOGGER.error('Syntax error:Scope Open Symbol {} with position = {} ignored (no corresponding closing bracket) in:\n {}'.format(token.word, token.pos, show(tokens)))
            else:
                (begin, end) = span
                (midtokens, midmetadata) = cleantokens(tokens[begin + 1:end], repkeep)
                newtokens += [tokens[tokenctr]] + midtokens + [tokens[end]]
                metadata += midmetadata
                tokenctr = end
        elif token.word == scope_close:
            settings.LOGGER.error('Syntax error: unexpected {} excountered with position {} in:\n {}'.format(token.word, token.pos, show(tokens)))
            newtokens.append(token)
        else:
            newtokens.append(token)
        tokenctr += 1
    return (newtokens, metadata)

#this is probably obsolete, it still uses Python2 syntax (print)


def checkline(line: str, newline: str, outfilename: str, lineno, logfile: TextIO):
    if checkpattern.search(newline) or pluspattern.search(newline):
        print(outfilename, lineno, 'suspect character', file=logfile)
        print('input=<{}>'.format(line[:-1]), file=logfile)
        print('output=<{}>'.format(newline), file=logfile)
        thecodes = str2codes(newline)
        print('charcodes=<{}>'.format(thecodes), file=logfile)


def purifytokens(tokens: List[Token]) -> List[Token]:
    result = [token for token in tokens if token.word not in illegalcleanedchatsymbols]
    return result


CleanedText = Union[List[Token], str]


def cleantext(utt: str, repkeep: bool, tokenoutput: bool = False) -> Tuple[CleanedText, Metadata]:
    '''

    :param utt:
    :param repkeep:
    :param tokenoutput:
    :return:
    '''
    newutt = robustness(utt)
    tokens = sastatok.sasta_tokenize(newutt)
    inwordlist = [t.word for t in tokens]
    intokenstrings = [str(token) for token in tokens]
    # print(space.join(intokenstrings))
    (newtokens, metadata) = cleantokens(tokens, repkeep)
    #remove symbol tokens that should not be there anymore
    newtokens = purifytokens(newtokens)
    resultwordlist = [t.word for t in newtokens]
    resultstring = smartjoin(resultwordlist)
    resultposlist = [t.pos for t in newtokens]
    newmeta1 = Meta('tokenisation', inwordlist, atype='list', source='CHAT/Tokenisation', backplacement=bpl_none)
    newmeta2 = Meta('cleanedtokenisation', resultwordlist, atype='list', source='CHAT/Tokenisation', backplacement=bpl_none)
    newmeta3 = Meta('cleanedtokenpositions', resultposlist, annotationposlist=resultposlist, atype='list', source='CHAT/Tokenisation', backplacement=bpl_none)
    metadata += [newmeta1, newmeta2, newmeta3]
    resultmetadata = metadata
    if tokenoutput:
        return (newtokens, resultmetadata)
    else:
        return (resultstring, resultmetadata)


def cleantokens(tokens: List[Token], repkeep: bool) -> Tuple[List[Token], Metadata]:
    newtokens = deepcopy(tokens)
    metadata = []

    (newtokens, nestingmetadata) = clearnesting(newtokens, repkeep)
    metadata += nestingmetadata

    for annotation in CHAT_Annotation.annotations:
        (newtokens, annotationmetadata) = annotation.apply(newtokens, repkeep)
        metadata += annotationmetadata
        tokenstrings = [str(token) for token in newtokens]
        # print(space.join(tokenstrings))

    return (newtokens, metadata)


def str2codes(str: str) -> List[Tuple[str, str]]:
    result = []
    for i in range(len(str)):
        curchar = str[i]
        curcode = hexformat.format(ord(str[i]))
        result.append((curchar, curcode))
    return (result)


def removesuspects(str: str) -> str:
    result1 = re.sub(checkpattern, space, str)
    result2 = re.sub(pluspattern1, r'\1', result1)
    result = re.sub(pluspattern2, r'\1', result2)
    return result


RobustnessTuple = Tuple[Pattern, str, str, str]

robustnessrules: List[RobustnessTuple] = [(re.compile(r'\u2026'), '\u2026', '...', 'Horizontal Ellipsis (\u2026, Unicode U+2026) replaced by a sequence of three Full Stops (..., Unicode U+002E) '),
                                          (re.compile('#'), '#', '', 'Number Sign (#, Unicode U+0023) removed'),
                                          #(re.compile('#'), '#', '(.)', 'Number Sign (#, Unicode U+0023) replaced by CHAT (short) pause code: (.)'),
                                          (re.compile(r'\[\+bch\]'), '[+bch]', '[+ bch]', 'Missing space'),
                                          (re.compile(r'\[\+trn\]'), '[+trn]', '[+ trn]', 'Missing space'),
                                          (re.compile(r'\[:(?![:\s])'), '[:', '[: ', 'Missing space'),
                                          (re.compile(r'(?<=\w)\+\.\.\.'), '+...', ' +...', 'Missing space'),
                                          (re.compile(r'\u2018'), '\u2018', "'", "Left Single Quotation Mark (\u2018. Unicode U+2018) replaced by Apostrophe ' (Unicode U+0027)"),
                                          (re.compile(r'\u2019'), '\u2019', "'", "Right Single Quotation Mark (\u2019, Unicode U+2019) replaced by Apostrophe ' (Unicode U+0027)"),
                                          (re.compile(r'\u201C'), '\u201C', '"', 'Left Double Quotation Mark (\u201C, Unicode U+201C) replaced by Quotation Mark (", Unicode U+0022)'),
                                          (re.compile(r'\u201D'), '\u201D', '"', 'Right Double Quotation Mark (\u201D, Unicode U+201D) replaced by Quotation Mark (", Unicode U+0022)')
                                          ]


def robustness(utt: str) -> str:
    newutt = utt
    for (regex, instr, outstr, msg) in robustnessrules:
        newnewutt = regex.sub(outstr, newutt)
        if newnewutt != newutt:
            settings.LOGGER.warning('{}. Interpreted <{}> as <{}> in <{}>'.format(msg, instr, outstr, utt))
        newutt = newnewutt
    return newutt


# checkpattern = re.compile(r'[][\(\)&%@/=><_0^~↓↑↑↓⇗↗→↘⇘∞≈≋≡∙⌈⌉⌊⌋∆∇⁎⁇°◉▁▔☺∬Ϋ123456789·\u22A5\u00B7\u0001\u2260\u21AB]')
checkpattern = re.compile(r'[][\(\)&%@/=><_^~↓↑↑↓⇗↗→↘⇘∞≈≋≡∙⌈⌉⌊⌋∆∇⁎⁇°◉▁▔☺∬Ϋ·\u22A5\u00B7\u0001\u2260\u21AB\u2039\u203A]')
# + should not occur except as compound marker black+board
# next one split up in order to do substitutions
pluspattern = re.compile(r'(\W)\+|\+(\W)')
pluspattern1 = re.compile(r'(\W)\+')
pluspattern2 = re.compile(r'\+(\W)')


def ispunctuation(wrd: str) -> bool:
    result = wrd in '.?!;,!<>'
    return result


def smartjoin(strings: List[str]) -> str:
    result = ''
    lstrings = len(strings)
    if lstrings > 0:
        for i in range(lstrings - 1):
            if ispunctuation(strings[i + 1]):
                result += strings[i]
            else:
                result += strings[i] + space
        result += strings[lstrings - 1]
    return result



'''
The *alpinoparsing* module provides functions for:

* parsing:

.. autofunction::  alpinoparsing::parse

* previewing a parse tree:

.. autofunction:: alpinoparsing::previewurl

'''

import logging
import re
import urllib.parse
import urllib.request

from lxml import etree  # type: ignore

from sastadev.memoize import memoize

#from sastadev.sastatypes import SynTree, URL

#from sastadev.config import settings.logger
#from sastadev.sastatypes import SynTree, URL

urllibrequestversion = urllib.request.__version__

alpino_special_symbols_pattern = r'[\[\]]'
alpino_special_symbols_re = re.compile(alpino_special_symbols_pattern)

gretelurl = 'https://gretel.hum.uu.nl/api/src/router.php/parse_sentence/'
#gretelurl = 'http://gretel.hum.uu.nl/api/src/router.php/parse_sentence/'
previewurltemplate = 'https://gretel.hum.uu.nl/ng/tree?sent={sent}&xml={xml}'
#previewurltemplate = 'http://gretel.hum.uu.nl/ng/tree?sent={sent}&xml={xml}'

emptypattern = r'^\s*$'
emptyre = re.compile(emptypattern)


def isempty(sent: str) -> bool:
    '''
    The function *isempty* checks whether the input string *sent* is the null string or
    consists of white space only.

    '''
    result = emptyre.match(sent) is not None
    return result


@memoize
#def parse(origsent: str, escape: bool = True) -> Optional[SynTree]:
def parse(origsent: str, escape: bool = True):
    '''
    The function *parse* invokes the alpino parser (over the internet, so an internet connection is required) to parse
    the string *origsent*.
    The parameter *escape* can be used to escape symbols that have a special meaning
    for Alpino. Its default value is *True*.

    This function is memoised, which might give unexpected results since the output type is mutable. Be careful if
    the same input string is parsed twice and the resulting objects should really be considered two different
    instances.!

    '''
    if isempty(origsent):
        return None
    if escape:
        sent = escape_alpino_input(origsent)
    else:
        sent = origsent
    encodedsent = urllib.parse.quote(sent)
    fullurl = gretelurl + encodedsent
    try:
        r1 = urllib.request.urlopen(fullurl)
    except urllib.request.HTTPError as e:
        logging.error('{}: parsing <{}> failed'.format(e, sent))
        return None
    except urllib.error.URLError as e:
        logging.error('{}: parsing <{}> failed'.format(e, sent))
        return None
    else:
        if 300 > r1.status >= 200:
            streebytes = r1.read()
            # print(streebytes.decode('utf8'))
            stree = etree.fromstring(streebytes)
            return stree
        else:
            logging.error('parsing failed:', r1.status, r1.reason, sent)
            return None

#def previewurl(stree: SynTree) -> URL:


def previewurl(stree):
    '''
    The function *previewurl* returns the URL to preview the input SynTree *stree* in the GreTEL application.
    Many browsers have a limit on the length of a URL (it is absurd that that is the
    case)
    and the  inferior MS Office software only accepts URLs with a maximum length of 255
    characters (prehistoric!), so the URLs generated by this function may be too long
    for such inferior software.

    '''
    sents = stree.xpath('.//sentence')
    if sents != []:
        sent = etree.tostring(sents[0])
    else:
        sent = ''
    xml = etree.tostring(stree)
    encodedsent = urllib.parse.quote(sent)
    encodedxml = urllib.parse.quote(xml)
    fullurl = previewurltemplate.format(sent=encodedsent, xml=encodedxml)
    return fullurl


def escape_alpino_input(instr: str) -> str:
    '''
    The function escape_alpino_input takes as input a string *str* and returns this string with symbols with a
    special meaning for Alpino escaped, in particular the square bracket symbols [ and ] used for bracketed input.
    :param instr:
    :return:
    '''
    result = ''
    for c in instr:
        if c == '[':
            newc = '\\['
        elif c == ']':
            newc = '\\]'
        else:
            newc = c
        result += newc
    return result


def test() -> None:
    while True:
        sent = input('sentence (ENTER to stop): ')
        if sent != '':
            stree = parse(sent)
            if stree is not None:
                print(stree)
        else:
            exit(0)


def test1() -> None:
    sent1 = 'Het slechte weer heeft al schade aangericht'
    stree = parse(sent1)
    thepreviewurl = previewurl(stree)
    with open('previewurl.txt', 'w', encoding='utf8') as outfile:
        print(thepreviewurl, file=outfile)


if __name__ == '__main__':
    test1()
    # test()



import logging

from sastadev import SD_DIR, alpinoparsing


class SastadevConfig:
    '''Class for keeping track of application configuration'''

    def __init__(self,
                 ALPINO_HOST: str = 'localhost',
                 ALPINO_PORT: int = 7001,
                 LOGGER=logging.getLogger('sastadev'),
                 SD_DIR: str = SD_DIR,
                 DATAROOT: str = '.',
                 PARSE_FUNC=alpinoparsing.parse
                 ):
        self.ALPINO_HOST = ALPINO_HOST
        self.ALPINO_PORT = ALPINO_PORT
        self.LOGGER = LOGGER
        self.SD_DIR = SD_DIR
        self.DATAROOT = DATAROOT
        self.PARSE_FUNC = PARSE_FUNC
        print('Configuration initiated.')


settings = SastadevConfig()



from typing import Callable, Dict, List

from sastadev.conf import settings
from sastadev.dedup import getposition
from sastadev.macros import expandmacros
from sastadev.metadata import (longrep, repeated, repeatedseqtoken, repetition,
                               substringrep)
from sastadev.sastatypes import SynTree
from sastadev.stringfunctions import string2list
from sastadev.tblex import asta_recognised_lexnode, asta_recognised_nounnode
from sastadev.treebankfunctions import (clausecats, find1, getattval,
                                        getnodeyield, getyield, showtns)

noun_xpath = './/node[%asta_noun%]'
expanded_noun_xpath = expandmacros(noun_xpath)

lex_path = './/node[%ASTA_LEX%]'
expanded_lex_xpath = expandmacros(lex_path)

astabijzinquerybase = './/node[%ASTA_Bijzin%]'
astabijzinquery = expandmacros(astabijzinquerybase)

astacoredelpvquery = './/node[%coredelpv%]'
expandedastacoredelpvquery = expandmacros(astacoredelpvquery)

dpancestorsquery = 'ancestor::node[@rel="dp"] | self::node[@rel="dp" or @rel="--"]'


def get_dupindex(stree: SynTree, cond: str) -> Dict[str, str]:
    dupindex = {}
    dupxpath = './/xmeta[{cond}]'.format(cond=cond)
    dupmetadatalist = stree.xpath(dupxpath)
    for meta in dupmetadatalist:
        if 'annotatedposlist' in meta.attrib and 'annotationposlist' in meta.attrib:
            keyliststr = meta.attrib['annotatedposlist']
            keylist = string2list(keyliststr)
            valueliststr = meta.attrib['annotationposlist']
            valuelist = string2list(valueliststr)
            if len(keylist) != len(valuelist):
                settings.LOGGER.error('Error in metadata: {} in sentence {}'.format(meta, getyield(stree)))
            else:
                for key, val in zip(keylist, valuelist):
                    dupindex[key] = val
    return dupindex


def asta_noun(
        stree: SynTree) \
        -> List[SynTree]:
    '''The function *asta_noun* uses the function *asta_x* with parameters *stree*,
    *expanded_noun_path* and the function *asta_recognised_nounnode*.

    The *expanded_noun_path* is the expansion of the macro **noun_path**.

    .. autofunction:: treebankfunctions::asta_recognised_nounnode

    '''
    results = asta_x(stree, expanded_noun_xpath, asta_recognised_nounnode)
    return results


def asta_lex(stree: SynTree) -> List[SynTree]:
    '''The function *asta_lex* uses the function *asta_x* with parameters *stree*,
    *expanded_lex_path* and the function *asta_recognised_nounnode*.

    The *expanded_lex_path* is the expansion of the macro **lex_path**.

    .. autofunction:: treebankfunctions::asta_recognised_lexnode
         :noindex:

    '''

    results = asta_x(stree, expanded_lex_xpath, asta_recognised_lexnode)
    return results


def old_asta_noun(stree: SynTree) -> List[SynTree]:
    theyield = getyield(stree)   # for debugging purposes
    thenodeyield = getnodeyield(stree)
    cond1 = '@value="{}" or @value="{}" or @value="{}" or @value="{}"'.format(repeated, repeatedseqtoken, longrep, substringrep)
    dupindex = get_dupindex(stree, cond1)
    cond2 = '@subcat="{}"'.format(repetition)
    allrepdupindex = get_dupindex(stree, cond2)
    revallrepdupindex = {val: key for key, val in allrepdupindex.items()}
    noun_nodes = stree.xpath(expanded_noun_xpath)
    # print(showtns(noun_nodes))

    clean_noun_nodes = noun_nodes

    # remove the nodes that should get it from this function
    clean_noun_nodes = [node for node in clean_noun_nodes if getattval(node, 'begin') not in allrepdupindex]

    # remove words not recognised as nouns; is dit nodig? Ja dit is nodig!!!
    clean_noun_nodes = [node for node in clean_noun_nodes if asta_recognised_nounnode(node)]
    # print(showtns(clean_noun_nodes))

    additional_nodes = []
    for key in dupindex:
        keynode = find1(stree, './/node[(@pt or @pos) and @begin="{}"]'.format(key))
        if keynode is not None:
            val = dupindex[key]
            valnode = find1(stree, './/node[(@pt or @pos) and @begin="{}"]'.format(val))
            if valnode is not None:
                if valnode in clean_noun_nodes:
                    additional_nodes.append(keynode)

    result = clean_noun_nodes + additional_nodes
    # print(showtns(result))
    return result


def verbleftof(node: SynTree, positions: List[str]) -> bool:
    nodebegin = getattval(node, 'begin')
    for position in positions:
        if int(position) < int(nodebegin):
            return True
    return False


def asta_delpv(stree: SynTree) -> List[SynTree]:
    coredelpvnodes = stree.xpath(expandedastacoredelpvquery)
    streeleaves = getnodeyield(stree)
    wwbegins = [getattval(node, 'begin') for node in streeleaves if getattval(node, 'pt') == 'ww']
    delpvnodes = [node for node in coredelpvnodes if node.xpath(dpancestorsquery) == [] or not (verbleftof(node, wwbegins))]
    return delpvnodes


def asta_x(stree: SynTree,
           xpathexpr: str,
           recognized_x_f: Callable[[SynTree], bool]) -> List[SynTree]:
    theyield = getyield(stree)   # for debugging purposes
    thenodeyield = getnodeyield(stree)
    cond1 = '@value="{}" or @value="{}" or @value="{}" or @value="{}"'.format(repeated, repeatedseqtoken, longrep, substringrep)
    dupindex = get_dupindex(stree, cond1)
    cond2 = '@subcat="{}"'.format(repetition)
    allrepdupindex = get_dupindex(stree, cond2)
    revallrepdupindex = {val: key for key, val in allrepdupindex.items()}
    x_nodes = stree.xpath(xpathexpr)
    # print(showtns(noun_nodes))

    clean_x_nodes = x_nodes

    # remove the nodes that should get it from this function
    clean_x_nodes = [node for node in clean_x_nodes if getattval(node, 'begin') not in allrepdupindex]

    # remove words not recognised as nouns; is dit nodig? Ja dit is nodig!!!
    clean_x_nodes = [node for node in clean_x_nodes if recognized_x_f(node)]
    # print(showtns(clean_noun_nodes))

    additional_nodes = []
    for key in dupindex:
        keynode = find1(stree, './/node[(@pt or @pos) and @begin="{}"]'.format(key))
        if keynode is not None:
            val = dupindex[key]
            valnode = find1(stree, './/node[(@pt or @pos) and @begin="{}"]'.format(val))
            if valnode is not None:
                if valnode in clean_x_nodes:
                    additional_nodes.append(keynode)

    result = clean_x_nodes + additional_nodes
    # print(showtns(result))
    return result


def getmluxnodes(mluxnodes, posnodes, dupinfo):
    resultnodes = []
    for node in mluxnodes:
        nodeposition = getposition(node)
        origpos = get_origpos(nodeposition, dupinfo)
        targetnode = find_node(origpos, posnodes)
        if targetnode is not None:
            resultnodes.append(node)
    return resultnodes


def get_origpos(nodeposition, dupinfo):
    newposition = nodeposition
    if newposition not in dupinfo.longdups:
        result = None
    else:
        while newposition in dupinfo.longdups:
            newposition = dupinfo.longdups[newposition]
        result = newposition
    return result


def find_node(position, nodes):
    results = [node for node in nodes if getposition(node) == position]
    lresults = len(results)
    if lresults == 0:
        result = None
    elif lresults == 1:
        result = results[0]
    else:
        settings.LOGGER.warning('Multiple nodes found for position {}: {}, in {}'.format(position, showtns(results), showtns(nodes)))
        result = results[0]
    return result


bijzin_xpath = './/node[%ASTA_Bijzin%]'
expanded_bijzin_xpath = expandmacros(bijzin_xpath)


def old_asta_bijzin(stree):
    candnodes = stree.xpath(expanded_bijzin_xpath)
    tops = stree.xpath('.//node[@cat="top"]')
    top = tops[0]
    done, resultingnodes = removehoofdzin(top, candnodes)
    return resultingnodes


def removerepetitions(ptnodes: List[SynTree], stree: SynTree) -> List[SynTree]:
    '''
    The function *removerepetitions* returns a list of nodes from *ptnodes* that are
    not marked in the metadata as repetitions.

    '''
    newptnodes = []
    for ptnode in ptnodes:
        ptnodeend = ptnode.attrib['begin'] if 'begin' in ptnode.attrib else None
        xmetaxpath = './/xmeta[@subcat="Repetition" and @annotatedposlist="[{}]"]'.format(ptnodeend)
        repmetas = stree.xpath(xmetaxpath)
        if repmetas == []:
            newptnodes.append(ptnode)
    return newptnodes


def asta_bijzin(stree: SynTree) -> List[SynTree]:
    '''
    The function *asta_bijzin* identifies *bijzinnen* in  *stree*. The term *bijzin*
    is used in ASTA in a slightly different way than usual.
    It usually means *subordinate clause*, but in ASTA it is equivalent to the English
    term *clause* (so it can be a main clause or a subordinate clause).

    The function finds each node that is a clause except for the left-most one. For
    example, in *wij gaan naar huis als hij ziek is* the leftmost clause starts at
    *wij* (and is not marked as a clause), and the second one starts with *als*,
    and that one is found as a *Bijzin*.

    If a main clause starts with a subordinate clause, then this subordinate clause is
    contained in the main clause: In e.g. *als hij ziek is gaan we naar huis* the
    subordinate clause is *als hij ziek is* and the main clause is *als hij ziek is
    gaan we naar huis*. We should mark *gaan* as the first word of a clause, but
    in the Alpino structure it is not the first word of a clause. For this reason we set
    the start of a clause
    that contains another clause with the same value for the *begin* attribute equal
    to the *end* attribute of the contained clause, so that a marking appears under the
    word *gaan*, the first word after the contained clause.

    The function launches an XPath query (*astabijzinquery*)  which crucially uses the
    macro *ASTA_Bijzin*. This query yields not only true clause nodes but also nodes
    of words that typically introduce clauses but have been parsed wrongly.
    These words may have been repeated, and if so, only one of them should be in the
    results, so the repetitions are removed from the results by means of the function
    *removerepetitions*:

    .. autofunction:: asta_queries::removerepetitions
    '''
    theyield = getyield(stree)
    clausenodes = stree.xpath(astabijzinquery)
    ptnodes = [n for n in clausenodes if 'pt' in n.attrib]
    okptnodes = removerepetitions(ptnodes, stree)
    trueclausenodes = [n for n in clausenodes if getattval(n, 'cat') in clausecats]
    # alternative 1
    # sortedclausenodes = sorted(trueclausenodes, key=lambda x: (int(getattval(x,'begin')), -int(getattval(x, 'end'))))
    # result = sortedclausenodes[1:] + okptnodes

    # alternative2 -follows the conventions for ASTA
    sortedclausenodes = sorted(trueclausenodes, key=lambda x: (int(getattval(x, 'begin')), int(getattval(x, 'end'))))
    if len(sortedclausenodes) > 1:
        cn0 = sortedclausenodes[0]
        cn1 = sortedclausenodes[1]
        if getattval(cn1, 'begin') == getattval(cn0, 'begin'):
            cn0end = getattval(cn0, 'end')
            newbegin = cn0end
            newokptnodexpath = '//node[@pt and @begin="{newbegin}"]'.format(newbegin=newbegin)
            newokptnode = find1(cn1, newokptnodexpath)
            result = sortedclausenodes[2:] + okptnodes
            if newokptnode is not None:
                result += [newokptnode]
        else:
            result = sortedclausenodes[1:] + okptnodes
    else:
        result = sortedclausenodes[1:] + okptnodes

    # ad hoc statement to ensure that there are no None matches should not happen anymore
    result = [el for el in result if el is not None]
    return result


def removehoofdzin(stree, clausenodes):
    resultingnodes = clausenodes
    done = False
    for child in stree:
        chatt = getattval(child, 'cat')
        if chatt in clausecats:
            if child in clausenodes:
                resultingnodes.remove(child)
                done = True
                return done, resultingnodes
            else:
                done = True
        if not done:
            done, resultingnodes = removehoofdzin(child, clausenodes)
    return done, resultingnodes



import pytest
from auchann.align_words import align_words

import sastadev.treebankfunctions as tbf
from sastadev import sasta_explanation


@pytest.mark.skip(reason='test code does not work')
def test():
    infullname = r"D:\Dropbox\jodijk\Utrecht\Projects\SASTADATA\Auris\outtreebanks\DLD07_corrected.xml"
    fulltreebank = tbf.getstree(infullname)
    treebank = fulltreebank.getroot()
    for tree in treebank:
        origutt = tbf.find1(tree, './/meta[@name="origutt"]/@value')
        cleanutt = tbf.find(tree, './/sentence/@value')
        explanationstr = sasta_explanation.finalmultiwordexplanation(tree)
        if explanationstr is not None:
            alignment = align_words(cleanutt, explanationstr)
        else:
            alignment = None
        if explanationstr is not None:
            print(
                f' Orig:{origutt}\nClean:{cleanutt}\n Expl:{explanationstr}\nAlign:{alignment}\n\n')


if __name__ == '__main__':
    test()



from typing import List

from lxml import etree

bpl_none, bpl_word, bpl_node, bpl_delete, bpl_indeze, bpl_extra_grammatical, bpl_wordlemma = tuple(range(7))
defaultpenalty = 10
defaultbackplacement = bpl_none

SASTA = 'SASTA'

xmlformat = '''
<xmeta name="{name}" type="{atype}" value= "{value}" annotationwordlist="{annotationwordlist}"
       annotationposlist="{annotationposlist}" annotatedwordlist="{annotatedwordlist}"
       annotatedposlist="{annotatedposlist}"  cat="{cat}" subcat="{subcat}" source="{source}"
       backplacement="{backplacement}" penalty="{penalty}"
/>\n
'''


class Meta:
    def __init__(self, name, value, annotationwordlist=[], annotationposlist=[], annotatedposlist=[],
                 annotatedwordlist=[], annotationcharlist=[], annotationcharposlist=[], annotatedcharlist=[],
                 annotatedcharposlist=[], atype='text', cat=None, subcat=None, source=None, penalty=defaultpenalty,
                 backplacement=defaultbackplacement):
        self.atype = atype
        self.name = name
        self.annotationwordlist = annotationwordlist if annotationwordlist != [] else value
        self.annotationposlist = annotationposlist
        self.annotatedwordlist = annotatedwordlist
        self.annotatedposlist = annotatedposlist
        self.annotationcharlist = annotationcharlist
        self.annotationcharposlist = annotationcharposlist
        self.annotatedcharlist = annotatedcharlist
        self.annotatedcharposlist = annotatedcharposlist
        self.value = value
        self.cat = cat
        self.subcat = subcat
        self.source = source
        self.penalty = penalty
        self.backplacement = backplacement
        self.fmstr = '<{}:type={}:annotationwordlist={}:annotationposlist={}:annotatedwordlist={}:annotatedposlist={}:value={}:cat={}:source={}>'
        self.xmlformat = xmlformat

    def __repr__(self):
        reprfmstr = 'Meta({},{},annotationwordlist={},annotationposlist={},annotatedposlist{},annotatedwordlist={},' \
                    ' atype={}, cat={}, subcat={}, source={}, penalty={}, backplacement={})'
        result = reprfmstr.format(repr(self.name), repr(self.value), repr(self.annotationwordlist), repr(self.annotationposlist),
                                  repr(self.annotatedposlist), repr(self.annotatedwordlist), repr(self.atype),
                                  repr(self.cat), repr(self.subcat), repr(self.source), repr(self.penalty), repr(self.backplacement))
        return result

    def __str__(self):
        frm = self.fmstr.format(self.name, self.atype, str(self.annotationwordlist),
                                str(self.annotationposlist), str(self.annotatedwordlist), str(self.annotatedposlist),
                                str(self.value), str(self.cat), str(self.source))
        return frm

    def toElement(self):
        # result = self.xmlformat.format(name=self.name, atype=self.atype, annotationwordlist=str(self.annotationwordlist),
        #                    annotationposlist=str(self.annotationposlist), annotatedwordlist=str(self.annotatedwordlist),
        #                    annotatedposlist=str(self.annotatedposlist), value=str(self.value), cat=str(self.cat),
        #                         subcat=self.subcat,  source=str(self.source), backplacement=self.backplacement,
        #                         penalty=self.penalty)

        result = etree.Element('xmeta', name=self.name, atype=self.atype, annotationwordlist=str(self.annotationwordlist),
                               annotationposlist=str(self.annotationposlist), annotatedwordlist=str(self.annotatedwordlist),
                               annotatedposlist=str(self.annotatedposlist), value=str(self.value), cat=str(self.cat),
                               subcat=str(self.subcat), source=str(self.source), backplacement=str(self.backplacement),
                               penalty=str(self.penalty))
        return result


def selectmeta(name, metadatalist):
    for meta in metadatalist:
        if meta.name == name:
            return meta
    return None


def mkSASTAMeta(token, nwt, name, value, cat, subcat=None, penalty=defaultpenalty, backplacement=defaultbackplacement):
    result = Meta(name, value, annotatedposlist=[token.pos],
                  annotatedwordlist=[token.word], annotationposlist=[nwt.pos],
                  annotationwordlist=[nwt.word], cat=cat, subcat=subcat, source=SASTA, penalty=penalty,
                  backplacement=backplacement)
    return result


Metadata = List[Meta]

#errormessages
filled_pause = "Filled Pause"
repeated = "Repeated word token"
repeatedseqtoken = "Word token of a repeated word token sequence"
repeatedjaneenou = "Repeated ja, nee, nou"
janeenou = "ja, nee or nou filled pause"
shortrep = 'Short Repetition'
longrep = 'Long Repetition'
intj = 'Interjection'
unknownword = 'Unknown Word'
unknownsymbol = 'Unknown Symbol'
substringrep = 'Substring repetition'
repetition = 'Repetition'
fstoken = 'Retraced token'
falsestart = 'Retracing with Correction'
insertion = 'Insertion'
smallclause = 'Small Clause Treatment'
tokenmapping = 'Token Mapping'
insertiontokenmapping = 'Insertion Token Mapping'



from collections import Counter

qidcolheader = 'id'
uttcolheader = 'uttids'
tab = '\t'
comma = ','
platinumheaderrows = 0


def read_referencefile(infilename, logfile):
    '''
    a reference file is tsv file which contains a header with at least two column headers (idcolheader, uttcolheader)
    :param infilename:
    :return: a dictionary with for each queryid a Counter for the utterance ids
    '''
    infile = open(infilename, 'r')
    rowctr = 0
    results = {}
    for line in infile:
        if rowctr == platinumheaderrows:
            rowstr = line.lower()
            rowlist = rowstr.split(tab)
            try:
                qidcol = rowlist.index(qidcolheader)
            except ValueError:
                print('Error reading reference file; no ID column header', infilename, file=logfile)
                exit(-1)
            try:
                uttcol = rowlist.index(uttcolheader)
            except ValueError:
                print('Error reading reference file: no uttids column header in', infilename, file=logfile)
                exit(-1)
        elif rowctr > platinumheaderrows:
            rowstr = line[:-1]
            rowlist = rowstr.split(tab)
            qid = rowlist[qidcol]
            utts = rowlist[uttcol]
            if utts == '':
                uttlist = []
            else:
                rawuttlist = utts.split(comma)
                uttlist = [uttid.strip() for uttid in rawuttlist]
            results[qid] = Counter(uttlist)
        rowctr += 1
    infile.close()
    return results



'''
Module to specify constant values that are used in the application
'''

# default locations of the data to train and test sastadev
bronzefolder = 'bronze'
formsfolder = 'forms'
indatafolder = 'indata'
intreebanksfolder = 'intreebanks'
loggingfolder = 'logging'
outtreebanksfolder = 'outtreebanks'
resultsfolder = 'results'
silverfolder = 'silver'
silverpermfolder = 'silverperm'



'''
The module *iedims* deals with diminutive forms ending in *-ie*. Alpino cannot deal with such diminutives,
but the function *getjeforms* finds for a given word ending in *-ie* what the corresponding diminutives ending in
*-je* are.
There can be many, e.g. *kassie* can correspond to *kasje* but also to *kastje*.

To our knowledge, there are no established orthographic rules for these forms. Therefore we should be robust to
several alternative ways of writing these forms, e.g.

* Kees: Keesie, Keessie, or Kesie?: all are accepted
* Huub: Huupie, Huuppie, or Hupie?: all are accepted
* wijf: wijfie, or wijffie?: we allow both
* feest: fesie, feesie, or feessie?: we require *feessie* or *feesie*
* dag: daggie, or dachie? : we require *daggie*
* truc: trucie, truuccie, truukkie, or  trukie?: we cannot deal with any at this moment
* maag: magie or maaggie, maachie?: we allow *magie* or *maaggie*
* vaas: vasie or vaassie?: both are allowed


The module has been tested against a gold reference of all words ending in ie(s) from OpenSonaR.
The module to use for a renewed test is tests.iedims_test.py, the file is tests.iediminutives.iedimsgold2.csv

.. autofunction:: iedims::getjeforms
'''


import re
from typing import List, Optional

from sastadev import lexicon

comma = ','
semicolon = ';'
bigfilesep = semicolon
slash = '/'
epsilon = ''
vertbar = '|'


def bracket(str: str) -> str:
    result = r'(' + str + r')'
    return result


def drop(pattern: str, chars: str) -> str:
    '''
    The function *drop* returns a string in which only those  characters of *pattern* are retained  if they do not
    occur in *chars*
    '''
    lpattern = [c for c in pattern]
    newlpattern = [c for c in lpattern if c not in chars]
    newpattern = epsilon.join(newlpattern)
    return newpattern


def regexor(stringlist: List[str]) -> str:
    '''
    The function *regexor* joins a list of strings by means of a vertical bar
    '''
    result = vertbar.join(stringlist)
    return result


def makecic1j() -> str:
    '''
    The function *makecic1j* creates a string that is a regular expression in which any of the consonants from c1 (
    =pfksg) is preceded by an arbitrary distinct consonant.

    '''
    parts = []
    for c in c1:
        if c not in '[]':
            newpart1 = drop(call, c1)
            newpart = '[' + newpart1 + ']' + c
            parts.append(newpart)
    pattern = regexor(parts)
    return pattern


begin = r'^(.*)'

vowel = r'[aeiouy]'
bvowel = bracket(vowel)
call = r'[bcdfghjklmnpqrstvwxz]'
bcall = bracket(call)
wb = '^'
callorwb = call + '|' + wb
bcallorwb = bracket(callorwb)
doublevowel = r'aa|ee|oo|uu'
diphthong = 'au|ei|eu|ie|ij|oe|ou|ui'
c1 = r'[pfksg]'
bc1 = bracket(c1)
ie = r'(ie)'
s = '(s?)'
end = r'$'
v4 = r'\4'
v3 = r'\3'
v2 = r'\2'
twovowels = doublevowel + '|' + diphthong
btwovowels = bracket(twovowels)
bdiphthong = bracket(diphthong)
cic1j = makecic1j()
bcic1j = bracket(cic1j)

#: * **Pattern**  vvssie: words ending in two vowels, two identicsl consonants from [pfksg] plus *ie*, optionally followed by *s*
#: * **Replacement**: degeminate the consonants and replace *ie* by *je* and *tje*
#: * **Examples**  meissie -> meisje; Keessie -> keessie; feessie -> feestje;  wijffie -> wijfje, lieffie -> liefje,
# boekkie -> boekje;
#:
vvssiepattern = begin + btwovowels + bc1 + v3 + r'ie' + s + end
vvssiere = re.compile(vvssiepattern, re.IGNORECASE)
# vvssiereplace: degeminate the consonants and replace *ie* by *je*
vvssiereplace = r'\1\2\3je\4'
# vvssiereplacet: degeminate the consonants and replce *ie* by *tje*
vvssiereplacet = r'\1\2\3tje\4'

#: * **Pattern** vvsie: the same as vvssie but now with a single consonant.
#: * **Replacement**: simply replace *íe* by *je* and *tje*
#: * **Examples**: koopie -> koopje, weekie -> weekje, feesie -> feestje, beesie -> beestje
#:
vvsiepattern = begin + btwovowels + bc1 + r'ie' + s + end
vvsiere = re.compile(vvsiepattern, re.IGNORECASE)
vvsiereplace = r'\1\2\3je\4'
vvsiereplacet = r'\1\2\3tje\4'


#: * **Pattern** CVCiCiie : ending in consonant (C), single vowel (V), two identical consonants (CiCi) and *ie(s)*
#: * **Replacement** VCije: degeminate the consonants and replace *ie* by *je*
#: * **Examples**: bakkie -> bakje ukkie -> ukje
#:
cvcicipattern = begin + bcallorwb + bvowel + bc1 + v4 + ie + s + end
cvcicire = re.compile(cvcicipattern, re.IGNORECASE)
cvcicireplace = r'\1\2\3\4je\6'
cvcicireplacet = r'\1\2\3\4tje\6'

#: * **Pattern** DCie : Diphthong followed by a single consonant and *ie(s)*
#: * **Replacement**  V2Cje: simply replace *ie* by *je*
#: * **Examples**: buikie > buikje wijfie -> wijfje buisie _-> buisje poepie -> poepje
#:
dciepattern = begin + bdiphthong + bc1 + r'ie' + s + end
dciere = re.compile(dciepattern, re.IGNORECASE)
dciereplace = r'\1\2\3je\4'

#: * **Pattern** CiC1jie: two different consonants followed by *ie(s)*
#: * **Replacement** C1C2je: simply replace *ie* by *je*
#: * **Examples**:# bankie -> bankje binkie -> binkje bloempie -> bloempje truckie -> truckje mamsie ->
# mamsje mensie -> mensje
cic1jiepattern = begin + bcic1j + r'ie' + s + end
cic1jiere = re.compile(cic1jiepattern, re.IGNORECASE)
cic1jiereplace = r'\1\2je\3'


# test
# testlist = [('bankie', 'bankje'), ('bloempie', 'bloempje'), ('truckie', 'truckje'), ('mensie', 'mensje')]
# localtest(cic1jiere, [cic1jre], testlist)

#: * **Pattern** chie: words ending in chie(s)
#: * **Replacement**: replace *ie* by *je* and by *tje*
#: * **Examples**: 'hachie', 'hachje'), ('vrachie', 'vrachtje'),  ('kuchie', 'kuchje')
chiepattern = begin + r'chie' + s + end
chiere = re.compile(chiepattern, re.IGNORECASE)
chiereplace = r'\1chje\2'
chiereplacet = r'\1chtje\2'

# test
# testlist = [('hachie', 'hachje'), ('vrachie', 'vrachtje'),  ('kuchie', 'kuchje')]
# localtest(chiere, [chiereplace, chiereplacet], testlist)


# Cvssie messie -> mesje; schoffie -> schoftje; etc. niet voor effie -> efje; actually not needed coverd by cvcicipattern
cvssiepattern = begin + bcall + bvowel + bc1 + v4 + r'ie' + s + end
cvssiere = re.compile(cvssiepattern, re.IGNORECASE)
cvssiereplace = r'\1\2\3\4je\5'
cvssiereplacet = r'\1\2\3\4tje\5'

# test
# testlist = [('messie', 'mesje'), ('zakkie', 'zakje'), ('schriffie', 'schriftje')]
# localtest(cvssiere, [cvssiereplace, cvssiereplacet], testlist)

#: * **Pattern** vcie: vowel + consonant _ *ie*
#: * **Replacement**: duplicate the vowel, replace *ie* by *je*
#: * **Examples**: ('slapie', 'slaapje'), ('rapie', 'raapje'), ('takie', 'taakje'), ('stafie', 'staafje'), ('magie','maagje'), ('vasie', 'vaasje')
#:
vciepattern = begin + bvowel + bcall + r'ie' + s + end
vciere = re.compile(vciepattern, re.IGNORECASE)
vciereplace = r'\1\2\2\3je\4'
# test
# testlist = [('slapie', 'slaapje'), ('rapie', 'raapje'), ('takie', 'taakje'), ('stafie', 'staafje'), ('magie', 'maagje'),
#            ('vasie', 'vaasje')]
# localtest(vciere, [vciereplace], testlist)

# getbase regexes:
#: * **Pattern** ngetjepattern: words ending in *ngetje(s)*
#: * **Replacement**: replace *ngetje* by *ng* and srop *s* if present
#: * **Examples**: ('tekeningetje','tekening'), ('tekeningetjes', 'tekening')
#:
ngetjepattern = begin + r'ngetje' + s + end     # tekeningetje
ngetjere = re.compile(ngetjepattern, re.IGNORECASE)
ngetjereplace = r'\1ng\2'

# testlist = [('tekeningetje','tekening'), ('tekeningetjes', 'tekening]
# localtest(ngetjere, [ngetjereplace], testlist)

#: * **Pattern** cicietje: words ending in two identical consonants followed by *etje(s)*
#: * **Replacement**: keep only on consonant and drop etjes(s)
#: * **Examples**: ('balletje','bal'), ('balletjes','bal')
#:
cicietjepattern = begin + bcall + v2 + r'etje' + s + end  # balletje
cicietjere = re.compile(cicietjepattern, re.IGNORECASE)
cicietjereplace = r'\1\2'

# testlist = [('balletje','bal'), ('balletjes','bal')]
# localtest(cicietjere, [cicietjereplace], testlist)

#: * **Pattern** cmpje: words ending in consonant plus *mpje(s)*
#: * **Replacement**: replace *mpje* by *m*, drop *s*
#: * **Examples**: ('armpje', 'arm'), ('armpjes', 'arm'), ('darmpje', 'darm'), ('darmpjes', 'darm')
#:
cmpjepattern = begin + bcall + r'mpje' + s + end  # darmpje, armpje
cmpjere = re.compile(cmpjepattern, re.IGNORECASE)
cmpjereplace = r'\1\2m'

# testlist = [('armpje', 'arm'), ('armpjes', 'arm'), ('darmpje', 'darm'), ('darmpjes', 'darm')]
# localtest(cmpjere, [cmpjereplace], testlist)

#: * **Pattern** vmpje: words ending in consonant + (two vowels or *e*) plus *mpje(s)*
#: * **Replacement**: replaces *mpje* by *m*, drop *s*
#: * **Examples**: ('bloempje','bloem'), ('bezempje', 'bezem'), ('bloempjes', 'bloem'), ('bezempjes', 'bezem')
#:
vmpjepattern = begin + bcall + bracket(regexor([btwovowels, 'e'])) + r'mpje' + s + end  # bloempje, bezempje
vmpjere = re.compile(vmpjepattern, re.IGNORECASE)
vmpjereplace = r'\1\2\3m'

# testlist = [('bloempje','bloem'), ('bezempje', 'bezem'), ('bloempjes', 'bloem'), ('bezempjes', 'bezem')]
# localtest(vmpjere, [vmpjereplace], testlist)


#: * **Pattern** nkjepattern: words ending in *nkje(s)*
#: * **Replacement**: replace *nkje* by *ng* and drop *s*
#: * **Examples**: ('koninkje','koning'), ('koninkjes','koning')
#:
nkjepattern = begin + r'nkje' + s + end   # koninkje
nkjere = re.compile(nkjepattern, re.IGNORECASE)
nkjereplace = r'\1ng'

# testlist = [('koninkje','koning'), ('koninkjes','koning')]
# localtest(nkjere, [nkjereplace], testlist)


#: * **Pattern** vivitje: words ending in two identical vowels followed by *tjes(s)*
#: * **Replacement**: keep one vowel, drop *tje*, drop *s*
#: * **Examples**: ('laatje','la'), ('laatjes','la')
#:
vivitjepattern = begin + bvowel + v2 + r'tje' + s + end  # laatje
vivitjere = re.compile(vivitjepattern, re.IGNORECASE)
vivitjereplace = r'\1\2'

# testlist = [('laatje','la'), ('laatjes','la')]
# localtest(vivitjere, [vivitjereplace], testlist)


#: * **Pattern** vivjtje: words ending in a diphthong followed by *tje(s)*
#: * **Replacement**: drop *tje*, drop *s* if present
#: * **Examples**: ('lelietje','lelie'), ('leitje', 'lei'), ('lelietjes','lelie'), ('leitjes', 'lei')
#:
vivjtjepattern = begin + bdiphthong + r'tje' + s + end   # lelietje, leitje
vivjtjere = re.compile(vivjtjepattern, re.IGNORECASE)
vivjtjereplace = r'\1\2'

# testlist = [('lelietje','lelie'), ('leitje', 'lei'), ('lelietjes','lelie'), ('leitjes', 'lei')]
# localtest(vivjtjere, [vivjtjereplace], testlist)


#: * **Pattern** je:  words ending in *je(s)*
#: * **Replacement**: drop *je*, drop *s* if present
#: * **Examples**: ('huisje','huis'), ('bakje', 'bak'), ('huisjes','huis'), ('bakjes', 'bak')
#:
jepattern = begin + r'je' + s + end  # huisje, bakje
jere = re.compile(jepattern, re.IGNORECASE)
jereplace = r'\1'

# testlist = [('huisje','huis'), ('bakje', 'bak'), ('huisjes','huis'), ('bakjes', 'bak')]
# localtest(jere, [jereplace], testlist)


voiceless = 'pst'
voiced = 'bzd'


def voicing(str: str) -> str:
    '''
    The function *voicing* replaces p by b, s by z and  t by d, for any other input string it returns the input string
    '''
    theindex = voiceless.find(str[0])
    if theindex >= 0:
        result = voiced[theindex]
    else:
        result = str
    return result


def getbaseinlexicon(dim: str) -> Optional[str]:
    '''
    The function *getbaseinlexicon* a lemma that occurs in the lexicon for the form *dim*, or None if no such lemma
    is found.

    '''
    result = None
    candidates = getbase(dim)
    for candidate in candidates:
        if lexicon.informlexicon(candidate):
            result = candidate
            break
    return result


def getbase(dim: str) -> List[str]:
    '''
    The function *getbase* computes a list of candidate lemmas for the input diminutive form (ending in *je*) *dim*.

    It makes use of various patterns and replacements:

    * .. autodata:: ngetjepattern
    * .. autodata:: cicietjepattern
    * .. autodata:: cmpjepattern
    * .. autodata:: vmpjepattern
    * .. autodata:: nkjepattern
    * .. autodata:: vivitjepattern
    * .. autodata:: vivjtjepattern
    * .. autodata:: jepattern

    '''
    results = []
    if ngetjere.match(dim):
        newresult = ngetjere.sub(ngetjereplace, dim)
        results.append(newresult)
    if cicietjere.match(dim):
        newresult = cicietjere.sub(cicietjereplace, dim)
        results.append(newresult)
    if cmpjere.match(dim):
        newresult = cmpjere.sub(cmpjereplace, dim)
        results.append(newresult)
    if vmpjere.match(dim):
        newresult = vmpjere.sub(vmpjereplace, dim)
        results.append(newresult)
    if nkjere.match(dim):
        newresult = nkjere.sub(nkjereplace, dim)
        results.append(newresult)
    if vivitjere.match(dim):
        newresult = vivitjere.sub(vivitjereplace, dim)
        results.append(newresult)
    if vivjtjere.match(dim):
        newresult = vivjtjere.sub(vivjtjereplace, dim)
        results.append(newresult)
    if jere.match(dim):
        newresult = jere.sub(jereplace, dim)
        results.append(newresult)
    return results


def getjeforms(ieform: str) -> List[str]:
    '''
    The function *getjeforms* when applied to a string  returns a list of strings that are diminutives ending in
    *-je* of which the lemma occurs in the lexicon. If no such string is found, it returns the empty list.

    It crucially makes use of the functions *getjeformsnolex* and *getbase*

    .. autofunction:: iedims::getjeformsnolex
    .. autofunction:: iedims::getbase

    '''
    results1 = getjeformsnolex(ieform)
    lemmas = []
    candpairs = []
    for result1 in results1:
        lemmas += getbase(result1)
        for lemma in lemmas:
            candpairs.append((result1, lemma))
    results = []
    for (form, lemma) in candpairs:
        if lexicon.known_word(lemma):
            # result = '[ @add_lex {} {} ]'.format(ieform, result)
            results.append(form)
    return results


def getjeformsnolex(ieform: str) -> List[str]:
    '''
    The function getjeformsnolex when applied to a string *ieform* returns a list of candidate
    diminutive strings ending in *-je* corresponding to  *ieform*. If no such candidates are found, it returns the
    empty list.

    The function applies a range of relevant regular expressions for *ie*-diminutives, and, if a match is found,
    it applies one or more appropriate replacements.

    The relevant regular expressions are:

    .. autodata:: cvcicipattern
    .. autodata:: dciepattern
    .. autodata:: vvsiepattern
    .. autodata:: vvssiepattern
    .. autodata:: vciepattern
    .. autodata:: chiepattern
    .. autodata:: cic1jiepattern

    For all patterns except for the last two mentioned also a voiced replacement is carried out.



    '''
    results = []
    if cvcicire.match(ieform):
        m = cvcicire.match(ieform)
        result = cvcicire.sub(cvcicireplace, ieform)
        results.append(result)
        result = cvcicire.sub(cvcicireplacet, ieform)
        results.append(result)
        bdg = voicing(m.group(4))
        result = cvcicire.sub(r'\1\2\3' + bdg + r'je\6', ieform)  # cluppie -> clubje, fessie -> fezje
        results.append(result)
    elif dciere.match(ieform):
        m = dciere.match(ieform)
        result = dciere.sub(dciereplace, ieform)
        results.append(result)
        bdg = voicing(m.group(3))
        result = dciere.sub(r'\1\2' + bdg + r'je\4', ieform)   # Huipie -> Huibje
        results.append(result)
    elif vvssiere.match(ieform):
        m = vvssiere.match(ieform)
        result = vvssiere.sub(vvssiereplace, ieform)
        results.append(result)
        result = vvssiere.sub(vvssiereplacet, ieform)
        results.append(result)
        bdg = voicing(m.group(3))
        result = vvssiere.sub(r'\1\2' + bdg + r'je\4', ieform)  # Huuppie -> Huubje
        results.append(result)
    elif vvsiere.match(ieform):
        m = vvsiere.match(ieform)
        result = vvsiere.sub(vvsiereplace, ieform)
        results.append(result)
        result = vvsiere.sub(vvsiereplacet, ieform)
        results.append(result)
        bdg = voicing(m.group(3))
        result = vvsiere.sub(r'\1\2' + bdg + r'je\4', ieform)  # Huupie -> Huubje
        results.append(result)
    elif vciere.match(ieform):  # must crucially occur after the previous two (otherwise beesie -> beeestje)
        m = vciere.match(ieform)
        result = vciere.sub(vciereplace, ieform)
        results.append(result)
        bdg = voicing(m.group(3))
        result = vciere.sub(r'\1\2\2' + bdg + r'je\4', ieform)   # Hupie -> Huubje
        results.append(result)
    elif chiere.match(ieform):
        # m = chiere.match(ieform)
        result = chiere.sub(chiereplace, ieform)
        results.append(result)
        result = chiere.sub(chiereplacet, ieform)
        results.append(result)
    elif cic1jiere.match(ieform):
        # m = cic1jiere.match(ieform)
        result = cic1jiere.sub(cic1jiereplace, ieform)
        results.append(result)
    else:
        results = []
    return results



import socket
from contextlib import contextmanager
from functools import lru_cache

from lxml import etree

import sastadev.conf
from sastadev.alpinoparsing import escape_alpino_input, isempty


class AlpinoSentenceParser:
    ''' Assumes a Alpino server is running on provided host:port,
    with assume_input_is_tokenized=off '''
    @contextmanager
    def connection(self):
        try:
            s = socket.create_connection((sastadev.conf.settings.ALPINO_HOST, sastadev.conf.settings.ALPINO_PORT))
            yield s
            s.close()
        except socket.error:
            raise

    def parse_sentence(self, sentence: str, buffer_size=8096) -> str:
        sentence = escape_alpino_input(sentence)
        with self.connection() as s:
            sentence += '\n\n'   # flag end of file
            s.sendall(sentence.encode('utf-8'))
            xml = b''
            while True:
                chunk = s.recv(buffer_size)
                if not chunk:
                    break
                xml += chunk
            return xml.decode('utf-8')


@lru_cache(maxsize=128)
def parse(sentence):
    ''' Wrapper for use in sastadev'''
    if isempty(sentence):
        return None
    alp = AlpinoSentenceParser()
    xml = alp.parse_sentence(sentence)
    return etree.fromstring(bytes(xml, encoding='utf-8'))



import re
import unicodedata
from typing import Any, Callable, List, Match, Optional, Sequence, Set

vertbar = '|'
space = ' '
hyphen = '-'
slash = '/'
tab = '\t'
comma = ','

csvre = "'[^']+'|[^,' ]+"  # for selecting nonempty tokens from a csvstring ; comma between single quotes is allowed
csvpat = re.compile(csvre)

wpat = r'^.*\w.*$'
wre = re.compile(wpat)
allhyphenspat = r'^-+$'
allhyphensre = re.compile(allhyphenspat)

barevowels = 'aeiouy'
aiguvowels = 'áéíóúý'
gravevowels = 'àèìòù\u00FD'
tremavowels = 'äëïöüÿ'
circumflexvowels = 'âêîôû\u0177'

consonants = 'bcdfghjklmnpqrstvwxz\u00E7'  # \u00E7 is c cedilla
dutch_base_vowels = barevowels + aiguvowels + gravevowels + tremavowels + circumflexvowels
vowels = dutch_base_vowels
dutch_base_diphthongs = ['aa', 'ee', 'ie', 'oo', 'uu', 'ij', 'ei', 'au', 'ou', 'ui', 'eu', 'oe']
dutch_y_diphthongs = ['y' + d for d in dutch_base_vowels] + [d + 'y' for d in
                                                             dutch_base_vowels]  # ryen gaat nog fout ye alleen samen nemen aan begin van woord
dutch_y_triphthongs = ['y' + d for d in dutch_base_diphthongs] + [d + 'y' for d in dutch_base_diphthongs]
dutch_trema_diphthongs = ['äa', "ëe", 'ïe', 'öo', 'üu', 'ëi']
dutch_diphthongs = dutch_base_diphthongs + dutch_y_diphthongs + dutch_trema_diphthongs
dutch_base_triphthongs = ['aai', 'eeu', 'ooi', 'oei']
dutch_y_tetraphthongs = ['y' + d for d in dutch_base_triphthongs]
dutch_triphthongs = dutch_base_triphthongs + dutch_y_triphthongs
dutch_tetraphthongs = dutch_y_tetraphthongs

hyphenprefixes = ['anti', 'contra', 'ex']

singlehyphenpat = r'(^[^-]+)-([^-]+)$'
singlehyphenre = re.compile(singlehyphenpat)

duppattern = r'(.)\1+'
dupre = re.compile(duppattern)

purechatxxxcodes = {'xxx', 'yyy', 'www'}
chatxxxcodes = purechatxxxcodes | {'xx'}


def pad(wrd: str, i: int, c: str = space) -> str:
    '''
    returns a right-justified string lengthened to length i and padded  with c
    '''
    if len(wrd) > i:
        result = wrd
    else:
        result = wrd.rjust(i, c)
    return result


def star(str: str) -> str:
    '''
    The function *star* takes a string str and returns it preceded by ( and followed by )*. Used to create
    regular expressions.
    '''
    return '({})*'.format(str)


def alt(strlist: Sequence[str], grouped: bool = True) -> str:
    '''
    The function alt takes as input a string or a list of strings and joins them into  a string separated by |.
    If grouped is True the resulting string is surrounded by round brackets, else not. Used to create regular expressions
    '''
    alts = '|'.join(strlist)
    if grouped:
        result = '({})'.format(alts)
    else:
        result = '{}'.format(alts)
    return result


def charrange(string: str) -> str:
    '''
    The function charrange take string str and surrounds it with square bracketys []. used to create regular expressions.
    '''
    return '[{}]'.format(string)


consonants_star = star(charrange(consonants))

syllableheadspat = alt([alt(dutch_tetraphthongs), alt(dutch_triphthongs), alt(dutch_diphthongs), alt(vowels)])
syllableheadsre = re.compile(syllableheadspat)

monosyllabicpat = r'^' + consonants_star + syllableheadspat + consonants_star + r'$'
monosyllabicre = re.compile(monosyllabicpat)


def barededup(word: str) -> str:
    '''
    The function barededup takes as input a string and
    returns a string in which  sequences of the same character are reduced to a single instance of this character
    and other characters are left unchanged. Example: 'vver' -> 'ver'
    '''
    result = dupre.sub(r'\1', word)
    return result


def deduplicate(word: str, inlexicon: Callable[[str], bool], exceptions: Set[str] = set()) -> List[str]:
    '''
    The function deduplicate takes as input a string word:

    * if it contains a sequence of duplicate characters, and
    * if it is not just a sequence of interpunction symbols, and
    * if it is not contained in the set of exceptions

    then

    * it checks whether the string with  the character sequence reduced to one character is a word
    according to the function *inlexicon*, and if so,  it adds this string to the result variable *newwords*
    * it checks whether the string with  the character sequence reduced to two characters is a word
    according to the function *inlexicon*, and if so,  it adds this string to the result variable *newwords*

    and then it returns the value of the result variable *newwords*
    '''
    newwords: List[str] = []
    if word in exceptions:
        newwords = []
    elif wre.match(word):  # we want to exclude tokens consisting of interpunction symbols only e.g  ---, --
        newword = dupre.sub(r'\1', word)
        if inlexicon(newword):
            newwords.append(newword)
        newword = dupre.sub(r'\1\1', word)
        if inlexicon(newword):
            newwords.append(newword)
    return newwords


def fullworddehyphenate(word: str, inlexicon: Callable[[str], bool]) -> List[str]:
    '''
    The function fullworddehyphenate takes as input a string *word*:

    Its purpose is to remove unnecessary hyphens from this word.

    * the hyphen can be a part of the word (*sergeant-majoor*), in which case it should not be removed.
    * it can also have been added to an existing word without hyphens (zie-ken-huis), in which case, it should be removed

    * it can also separate a possibly mispronounced prefix repetition of the prefix of word (e.g., ver-verkoopt or vver-verkoopt), in which case the hyphen and the prefix should be removed.


    To that end,

    * it applies the function *dehyphenate* to *word*. If this yields a result that is an existing word according to the function *inlexicon*, then this result is added to the result variable *newtokens*

    * if newtokens is still the empty list after this, it applies the function *delhyphenprefix* to word.  If this yields a result that is an existing word according to the function *inlexicon*, then this result is added to the result variable *newtokens*.

    and then it returns the result variable newtokens.

    The functions *dehyphenate* and *delhyphenprefix* are described here:

    * .. autofunction:: stringfunctions::dehyphenate
    * .. autofunction:: stringfunctions::delhyphenprefix


    '''
    newtokens = []
    newwords = dehyphenate(word)
    newwordset = set(newwords)
    for newword in newwordset:
        if inlexicon(newword):
            newtokens.append(newword)
    if newtokens == []:
        newwords = delhyphenprefix(word, inlexicon)
        newwordset = set(newwords)
        for newword in newwordset:
            newtokens.append(newword)
    return newtokens


def delhyphenprefix(word: str, inlexicon: Callable[[str], bool]) -> List[str]:
    '''
    The function *delphyphenprefix* takes as input a string *word*, splits it into a prefix and a mainword
    based on the first occurring hyphen, and  then:

    * if the prefix is a prefix that normally occurs with a hyphen and the mainword is an existing word according to the function *inlexicon* (e.g. *ex-vrouw*), the result variable is set to the empty list;
    * if the mainwords starts with the prefix and the main word is an existing word according to the function *inlexicon* (e.g. *ver-verkoop*),  then the result variable is set to [mainword];
    * if the prefix and the mainword are both existing words, the result variable is set to [];
    * if the mainwords starts with barededeup(prefix) and the main word is an existing word according to the function *inlexicon* (e.g. *vver-verkoop*),  then the result variable is set to [mainword]

     and then it returns the value of the result variable *result*.

    '''
    m = singlehyphenre.match(word)
    if m is not None:
        prefix = m.group(1)
        mainword = m.group(2)
        mwinlex = inlexicon(mainword)
        pfinlex = inlexicon(prefix)
        deduppf = barededup(prefix)
        if prefix in hyphenprefixes and mwinlex:  # the word starts wit ha known prefix that uses hyphen such as ex (ex-vrouw)
            result = []
        elif mainword.startswith(prefix) and mwinlex:  # this is the core case  e.g. ver-verkoop
            result = [mainword]
        elif pfinlex and mwinlex:  # for compounds with a hyphen: kat-oorbellen, generaal-majoor and for tennis-baan(?)
            result = []
        elif mainword.startswith(deduppf) and mwinlex:  # vver-verkoop
            result = [mainword]
        else:
            result = []
    else:
        result = []
    return result


def allhyphens(word: str) -> Optional[Match]:
    '''
    The fucntion allhyphens checks whether the strin gword consist completely of hyphens
    '''
    result = allhyphensre.match(word)
    return result


def dehyphenate(word: str) -> List[str]:
    '''
    The function dehyphenate takes as input a string and returns  a list of strings with all possible ways
    of removing hyphens in this string
    Examples:

    * dehyphenate('zie-ken-huis') = ['zie-ken-huis', 'zieken-huis', 'zie-kenhuis', 'ziekenhuis']
    * dehyphenate('ziekenhuis') = ['ziekenhuis']
    * dehyphenate('---') = ['---']  (a string consisting only of hyphens remains unchanged

    '''
    results = []
    if len(word) == 0:
        results = ['']
    elif allhyphens(word):
        results = [word]
    else:
        head = word[0:1]
        tail = word[1:]
        if head == hyphen:
            # newresult = head + tail
            # results.append(newresult)
            rightresults = dehyphenate(tail)
            for rightresult in rightresults:
                newresult = head + rightresult
                results.append(newresult)
                newresult = rightresult
                results.append(newresult)
        else:
            tailresults = dehyphenate(tail)
            for tailresult in tailresults:
                newresult = head + tailresult
                results.append(newresult)
    return results


def isconsonant(char: str) -> bool:
    '''
    The function isconsonant checks whether the input string char is a consonant.
    '''
    if len(char) != 1:
        result = False
    elif char.lower() in consonants:
        result = True
    else:
        result = False
    return result


def isvowel(char: str) -> bool:
    '''
    The function isvowel checks whether the input string char is a vowel.
    '''
    if len(char) != 1:
        result = False
    elif char.lower() in vowels:
        result = True
    else:
        result = False
    return result


def endsinschwa(word: str) -> bool:
    '''
    The function endinschwa checks whether the string word ends in schwa.
    '''
    # Ce of Vie of ije or ë
    if word[-3:] == 'ije':
        result = True
    elif word[-2:] == 'ie' and isvowel(word[-3:-2]):
        result = True
    elif word[-1:] == "ë":
        result = True
    elif word[-1:] == 'e' and isconsonant(word[-2:-1]):
        result = True
    else:
        result = False
    return result


def isdiphthong(d: str) -> bool:
    '''
    The function isdiphthong checks whether the string d is a diphtong in Dutch.
    '''
    result = d in dutch_diphthongs
    return result


def istriphthong(d: str) -> bool:
    '''
    The function istriphthong checks whether the string d is a triphtong in Dutch.
    '''
    result = d in dutch_triphthongs
    return result


def monosyllabic(word: str) -> bool:
    '''
    The function monosyllabic checks whether the string word cosnist of one syllable.
    '''
    result = monosyllabicre.match(word) is not None
    return result


def accentaigu(word: str) -> List[str]:
    '''
    The function accentaigu turns all vowels of a word into their varinats with an accent aigu, in all possible combinations.
    Example: accentaigu('aap') == ['aap', 'aáp', 'áap', 'ááp']
    '''
    if len(word) == 0:
        results = ['']
    elif isvowel(word[0]):
        restresults = accentaigu(word[1:])
        results1 = [word[0] + wrest for wrest in restresults]
        results2 = [aigu(word[0]) + wrest for wrest in restresults]
        results = results1 + results2
    else:
        restresults = accentaigu(word[1:])
        results = [word[0] + wrest for wrest in restresults]
    return results


def aigu(c: str) -> str:
    '''
    The function aigu turns the string c into a variant with an accent aigu.
    Only useful for single character strings taht are contained in barevowels
    :param c:
    :return:
    '''
    theindex = barevowels.find(c)
    result = aiguvowels[theindex]
    return result


def testcondition(condition, word):
    if condition(word):
        print('OK:{}'.format(word))
    else:
        print('NO:{}'.format(word))


def test():
    monosyllabicwords = ['baai', 'eeuw', 'mooi', 'aap', 'deed', 'Piet', 'noot', 'duut', 'rijd', 'meid', 'rauw', 'koud',
                         'buit', 'reuk', 'boer', 'la', 'de', 'hik', 'dop', 'dut',
                         'yell', 'ry', 'Händl', 'Pëtr', 'bït', 'Köln', 'Kür', 'Tÿd']
    disyllabicwords = ['baaien', 'eeuwen', 'mooie', 'aapje', 'deden', 'Pietje', 'noten', 'dut', 'rijden', 'meiden',
                       'rauwe', 'koude', 'buitje', 'reuken', 'boeren', 'laden', 'dender',
                       'hikken', 'doppen', 'dutten', 'yellen', 'ryen', 'Händler', 'Pëtri', 'bïty', 'Kölner', 'Kürer',
                       'Tÿding', 'naäap', 'meeëten', 'ciën', 'coöp']

    for word in monosyllabicwords:
        testcondition(monosyllabic, word.lower())
    for word in disyllabicwords:
        testcondition(monosyllabic, word.lower())

    for word in monosyllabicwords + disyllabicwords:
        ms = syllableheadsre.finditer(word)
        print(word, end=' -- ')
        for m in ms:
            print(m.group(0), end=', ')
        print('')


def nono(inval: Any) -> bool:
    result = (inval is None) or (inval == 0) or (inval == []) or (inval == '')
    return result


def nonnull(inval: Any) -> bool:
    result = not (nono(inval))
    return result


def allconsonants(inval: str) -> bool:
    '''
    The function allconsonants checks whether all characters in the string inval are consonants
    '''
    result = all([isconsonant(c) for c in inval])
    return result


def string2list(liststr: str, quoteignore=False) -> List[str]:
    '''
    The function string2list turns a string surrounded by [ ] into a list by splitting it on a comma
    Examples:
    * "[1,2,3]" becomes ['1', '2', '3']
    * "[]" becomes []
    * "[ ]" becomes [" "]
    * "['Jan', 'Piet']" becomes ["'Jan'", "'Piet'"] if quoteignore = False
    * "['Jan', 'Piet']" becomes ['Jan', 'Piet'] if quoteignore = True
    * comma's between single quotes are allowed

    '''
    if liststr is None or len(liststr) == 2:
        return []
    elif liststr[0] == '[' and liststr[-1] == ']':
        core = liststr[1:-1]
        parts = csvpat.findall(core)
        strippedparts = [part.strip() for part in parts]
        if quoteignore:
            cleanparts1 = [part.strip("'") for part in strippedparts]
            cleanparts = [part.strip('"') for part in cleanparts1]
        else:
            cleanparts = strippedparts
        return cleanparts
    else:
        return []


def realwordstring(w: str) -> bool:
    '''
    The function realwordstrin gcheck whether the string w @@ to be extended@@

    '''
    if len(w) != 1:
        result = True
    else:
        result = not unicodedata.category(w).startswith('P')
    return result


if __name__ == '__main__':
    test()



class TokenListMD():
    def __init__(self, tokens, metadata):
        self.tokens = tokens
        self.metadata = metadata

    def __repr__(self):
        tokstr = repr(self.tokens)
        mdstr = repr(self.metadata)
        result = 'TokenListMD({},{})'.format(tokstr, mdstr)
        return result


class TokenMD():
    def __init__(self, token, metadata):
        self.token = token
        self.metadata = metadata

    def __repr__(self):
        tokstr = repr(self.token)
        mdstr = repr(self.metadata)
        result = 'TokenMD({},{})'.format(tokstr, mdstr)
        return result


def mdlist2listmd(mdlist):
    newtokens = []
    mergedmetadata = []
    for tokenmd in mdlist:
        newtokens.append(tokenmd.token)
        mergedmetadata += tokenmd.metadata
    result = TokenListMD(newtokens, mergedmetadata)
    return result
